"""Plotting functions."""

import logging
from itertools import zip_longest

import numpy as np
import pandas as pd
import seaborn as sns
import nibabel as nib
import matplotlib
import matplotlib.pyplot as plt
from matplotlib import patches
from matplotlib.path import Path
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from nilearn.plotting import plot_markers, plot_img_on_surf

from osl_dynamics import files
from osl_dynamics.array_ops import get_one_hot
from osl_dynamics.utils.misc import override_dict_defaults
from osl_dynamics.utils.topoplots import Topology
from osl_dynamics.utils.parcellation import Parcellation, parcel_vector_to_voxel_grid


_logger = logging.getLogger("osl-dynamics")

# Suppress matplotlib warnings
logging.getLogger("matplotlib.category").setLevel(logging.ERROR)


def set_style(params):
    """Sets matplotlib's style.

    Wrapper for `plt.rcParams.update
    <https://matplotlib.org/stable/tutorials/introductory/customizing.html>`_.
    List of parameters can be found `here
    <https://matplotlib.org/stable/api/matplotlib_configuration_api.html\
    #matplotlib.rcParams>`_.

    Parameters
    ----------
    params : dict
        Dictionary of style parameters to update.

    Examples
    --------
    Make labels and linewidth larger::

        plotting.set_style({
            "axes.labelsize": 16,
            "xtick.labelsize": 14,
            "ytick.labelsize": 14,
            "legend.fontsize": 14,
            "lines.linewidth": 3,
        })
    """
    plt.rcParams.update(params)


def create_figure(*args, **kwargs):
    """Creates matplotlib figure and axes objects.

    Parameters
    ----------
    fig_kwargs
        Arguments to pass to `plt.subplots <https://matplotlib.org/stable/api\
        /_as_gen/matplotlib.pyplot.subplots.html>`_.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure.
    ax : array of plt.axes
        Array of axes (or single axis).
    """
    fig, ax = plt.subplots(*args, **kwargs)
    return fig, ax


def show(tight_layout=False):
    """Displays all figures in memory.

    Wrapper for `plt.show <https://matplotlib.org/stable/api/_as_gen/\
    matplotlib.pyplot.show.html>`_.

    Parameters
    ----------
    tight_layout : bool, optional
        Should we call :code:`plt.tight_layout()`?
    """
    if tight_layout:
        plt.tight_layout()
    plt.show()


def save(fig, filename, tight_layout=False):
    """Save and close a figure.

    Parameters
    ----------
    fig : plt.figure
        Matplotlib figure object.
    filename : str
        Output filename.
    tight_layout : bool, optional
        Should we call :code:`fig.tight_layout()`?
    """
    _logger.info(f"Saving {filename}")
    if tight_layout:
        fig.tight_layout()
    fig.savefig(filename)
    close(fig)


def close(fig=None):
    """Close a figure.

    Parameters
    ----------
    fig : plt.figure, optional
        Figure to close. Defaults to all figures.
    """
    if fig is None:
        fig = "all"
    plt.close(fig)


def rough_square_axes(n_plots):
    """Get the most square axis layout for n_plots.

    Given :code:`n_plots`, find the side lengths of the rectangle which gives
    the closest layout to a square grid of axes.

    Parameters
    ----------
    n_plots : int
        Number of plots to arrange.

    Returns
    -------
    short : int
        Number of axes on the short side.
    long : int
        Number of axes on the long side.
    empty : int
        Number of axes left blank from the rectangle.
    """
    long = np.floor(n_plots**0.5).astype(int)
    short = np.ceil(n_plots**0.5).astype(int)
    if short * long < n_plots:
        short += 1
    empty = short * long - n_plots
    return short, long, empty


def get_colors(n, colormap="magma"):
    """Produce equidistant colors from a matplotlib colormap.

    Given a matplotlib colormap, produce a series of RGBA colors which are
    equally spaced by value. There is no guarantee that these colors will be
    perceptually uniformly distributed and with many colors will likely be
    extremely close.

    Parameters
    ----------
    n : int
        The number of colors to return.
    colormap : str, optional
        The name of a matplotlib colormap.

    Returns
    -------
    colors: list of tuple of float
        Colors in RGBA format.

    Note
    ----
    :code:`alpha=1.0` for all colors.
    """
    colormap = plt.get_cmap(colormap)
    colors = [colormap(1 * i / n) for i in range(n)]
    return colors


def plot_line(
    x,
    y,
    labels=None,
    legend_loc=1,
    errors=None,
    x_range=None,
    y_range=None,
    x_label=None,
    y_label=None,
    title=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Basic line plot.

    Parameters
    ----------
    x : list of np.ndarray
        x-ordinates.
    y : list of np.ndarray
        y-ordinates.
    labels : list of str, optional
        Legend labels for each line.
    legend_loc : int, optional
        Matplotlib legend location identifier. Default is top right.
    errors : list with 2 items, optional
        Min and max errors.
    x_range : list, optional
        Minimum and maximum for x-axis.
    y_range : list, optional
        Minimum and maximum for y-axis.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    plot_kwargs : dict, optional
        Arguments to pass to the `ax.plot <https://matplotlib.org/stable\
        /api/_as_gen/matplotlib.axes.Axes.plot.html>`_ method.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if len(x) != len(y):
        raise ValueError("Different number of x and y arrays given.")

    if x_range is None:
        x_range = [None, None]

    if y_range is None:
        y_range = [None, None]

    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        else:
            if len(labels) != len(x):
                raise ValueError("Incorrect number of lines or labels passed.")
        add_legend = True
    else:
        labels = [None] * len(x)
        add_legend = False

    if errors is None:
        errors_min = [None] * len(x)
        errors_max = [None] * len(x)
    elif len(errors) != 2:
        raise ValueError(
            "Errors must be errors=[[y_min1, y_min2,...], [y_max1, y_max2,..]]."
        )
    elif len(errors[0]) != len(x) or len(errors[1]) != len(x):
        raise ValueError("Incorrect number of errors passed.")
    else:
        errors_min = errors[0]
        errors_max = errors[1]

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot lines
    for x_data, y_data, label, e_min, e_max in zip(
        x, y, labels, errors_min, errors_max
    ):
        ax.plot(x_data, y_data, label=label, **plot_kwargs)
        if e_min is not None:
            ax.fill_between(x_data, e_min, e_max, alpha=0.3)

    # Set axis range
    ax.set_xlim(x_range[0], x_range[1])
    ax.set_ylim(y_range[0], y_range[1])

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Add a legend
    if add_legend:
        ax.legend(loc=legend_loc)

    # Save figure
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_scatter(
    x,
    y,
    labels=None,
    legend_loc=1,
    errors=None,
    x_range=None,
    y_range=None,
    x_label=None,
    y_label=None,
    title=None,
    markers=None,
    annotate=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Basic scatter plot.

    Parameters
    ----------
    x : list of np.ndarray
        x-ordinates.
    y : list of np.ndarray
        y-ordinates.
    labels : list of str, optional
        Legend labels for each line.
    legend_loc : int, optional
        Matplotlib legend location identifier. Default is top right.
    errors : list, optional
        Error bars.
    x_range : list, optional
        Minimum and maximum for x-axis.
    y_range : list, optional
        Minimum and maximum for y-axis.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    markers : list of str, optional
        Markers to used for each set of data points.
    annotate : List of array like objects, optional
        Annotation for each data point for each set of data points.
    plot_kwargs : dict, optional
        Arguments to pass to the `ax.scatter <https://matplotlib.org/stable\
        /api/_as_gen/matplotlib.axes.Axes.scatter.html>`_ method.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if len(x) != len(y):
        raise ValueError("Different number of x and y arrays given.")

    if x_range is None:
        x_range = [None, None]

    if y_range is None:
        y_range = [None, None]

    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        else:
            if len(labels) != len(x):
                raise ValueError("Incorrect number of data points or labels passed.")
        add_legend = True
    else:
        labels = [None] * len(x)
        add_legend = False

    if errors is None:
        errors = [None] * len(x)

    if markers is not None:
        if len(markers) != len(x):
            raise ValueError("Incorrect number of data points or markers passed.")
    else:
        markers = [None] * len(x)

    if annotate is not None:
        if len(annotate) != len(x):
            raise ValueError("Incorrect number of data points or annotates passed.")
    else:
        annotate = [None] * len(x)

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    if len(x) > 10:
        colors = get_colors(len(x), colormap="tab20")
    else:
        colors = get_colors(len(x), colormap="tab10")

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot data
    for i in range(len(x)):
        ax.scatter(
            x[i],
            y[i],
            label=labels[i],
            marker=markers[i],
            color=colors[i],
            **plot_kwargs,
        )
        if errors[i] is not None:
            ax.errorbar(x[i], y[i], yerr=errors[i], fmt="none", c=colors[i])
        if annotate[i] is not None:
            for j, txt in enumerate(annotate[i]):
                ax.annotate(txt, (x[i][j], y[i][j]))

    # Set axis range
    ax.set_xlim(x_range[0], x_range[1])
    ax.set_ylim(y_range[0], y_range[1])

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Add a legend
    if add_legend:
        ax.legend(loc=legend_loc)

    # Save figure
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_hist(
    data,
    bins,
    labels=None,
    legend_loc=1,
    x_range=None,
    y_range=None,
    x_label=None,
    y_label=None,
    title=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Basic histogram plot.

    Parameters
    ----------
    data : list of np.ndarray
        Raw data to plot (i.e. non-histogramed data).
    bins : list of int
        Number of bins for each item in data.
    labels : list of str, optional
        Legend labels for each line.
    legend_loc : int, optional
        Matplotlib legend location identifier. Default is top right.
    x_range : list, optional
        Minimum and maximum for x-axis.
    y_range : list, optional
        Minimum and maximum for y-axis.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    plot_kwargs : dict, optional
        Arguments to pass to the `ax.hist <https://matplotlib.org/stable\
        /api/_as_gen/matplotlib.axes.Axes.hist.html>`_ method. Defaults to
        :code:`{"histtype": "step"}`.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if len(data) != len(bins):
        raise ValueError("Different number of bins and data.")

    if x_range is None:
        x_range = [None, None]

    if y_range is None:
        y_range = [None, None]

    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        else:
            if len(labels) != len(data):
                raise ValueError("Incorrect number of labels or data passed.")
        add_legend = True
    else:
        labels = [None] * len(data)
        add_legend = False

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}
    default_plot_kwargs = {"histtype": "step"}
    plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot histograms
    for d, b, l in zip(data, bins, labels):
        ax.hist(d, bins=b, label=l, **plot_kwargs)

    # Set axis range
    ax.set_xlim(x_range[0], x_range[1])
    ax.set_ylim(y_range[0], y_range[1])

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Add a legend
    if add_legend:
        ax.legend(loc=legend_loc)

    # Save the figure if a filename has been pass
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_bar_chart(
    counts,
    x=None,
    x_range=None,
    y_range=None,
    x_label=None,
    y_label=None,
    title=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Bar chart plot.

    Parameters
    ----------
    counts : list of np.ndarray
        Data to plot.
    x : list or np.ndarray, optional
        x-values for counts.
    x_range : list, optional
        Minimum and maximum for x-axis.
    y_range : list, optional
        Minimum and maximum for y-axis.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    plot_kwargs : dict, optional
        Arguments to pass to the `ax.bar <https://matplotlib.org/stable\
        /api/_as_gen/matplotlib.axes.Axes.bar.html>`_ method.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if x is None:
        x = range(1, len(counts) + 1)
    elif len(x) != len(counts):
        raise ValueError("Incorrect number of x-values or counts passed.")
    else:
        x = [str(xi) for xi in x]

    if x_range is None:
        x_range = [None, None]

    if y_range is None:
        y_range = [None, None]

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot bar chart
    ax.bar(x, counts, **plot_kwargs)

    # Set axis range
    ax.set_xlim(x_range[0], x_range[1])
    ax.set_ylim(y_range[0], y_range[1])

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Save the figure if a filename has been pass
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_gmm(
    data,
    amplitudes,
    means,
    stddevs,
    bins=50,
    legend_loc=1,
    x_range=None,
    y_range=None,
    x_label=None,
    y_label=None,
    title=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot a two component Gaussian mixture model.

    Parameters
    ----------
    data : np.ndarray
        Raw data to plot as a histogram.
    amplitudes : np.ndarray
        Amplitudes of each Gaussian component.
        Mixture weights scaled by mixture covariances.
    means : np.ndarray
        Mean of each Gaussian component.
    stddevs : np.ndarray
        Standard deviation of each Gaussian component.
    bins : list of int, optional
        Number of bins for the historgram.
    legend_loc : int, optional
        Position for the legend.
    x_range : list, optional
        Minimum and maximum for x-axis.
    y_range : list, optional
        Minimum and maximum for y-axis.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if x_range is None:
        x_range = [None, None]

    if y_range is None:
        y_range = [None, None]

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot histogram
    ax.hist(data, bins=bins, histtype="step", density=True, color="tab:blue")

    # Plot Gaussian components
    x = np.arange(min(data), max(data), (max(data) - min(data)) / bins)
    y1 = amplitudes[0] * np.exp(-((x - means[0]) ** 2) / (2 * stddevs[0] ** 2))
    y2 = amplitudes[1] * np.exp(-((x - means[1]) ** 2) / (2 * stddevs[1] ** 2))
    ax.plot(x, y1, label="Off", color="tab:orange")
    ax.plot(x, y2, label="On", color="tab:green")
    ax.plot(x, y1 + y2, color="tab:red")

    # Set axis range
    if not any(r is None for r in x_range):
        ax.set_xlim(x_range[0], x_range[1])
    if not any(r is None for r in y_range):
        ax.set_ylim(y_range[0], y_range[1])

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Add legend
    if legend_loc is not None:
        ax.legend(loc=legend_loc)

    # Save the figure if a filename has been pass
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_violin(
    data,
    x=None,
    x_label=None,
    y_label=None,
    title=None,
    fig_kwargs=None,
    sns_kwargs=None,
    ax=None,
    filename=None,
):
    """Violin plot.

    Parameters
    ----------
    data : list of np.ndarray
        Data to plot.
    x : list or np.ndarray, optional
        x-values for data.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    sns_kwargs : dict, optional
        Arguments to pass to :code:`sns.violinplot()`.
    ax : matplotlib.axes.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if x is None:
        x = np.arange(len(data)) + 1
    elif len(x) != len(data):
        raise ValueError("Incorrect number of x-values or data passed.")
    else:
        x = [str(xi) for xi in x]

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if sns_kwargs is None:
        sns_kwargs = {}

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot violins
    x = np.concatenate([[x_] * len(y) for x_, y in zip(x, data)])
    y = data.flatten()
    ax = sns.violinplot(x=x, y=y, hue=x, ax=ax, legend=False, **sns_kwargs)

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)

    # Save the figure if a filename has been pass
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_time_series(
    time_series,
    n_samples=None,
    y_tick_values=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot a time series with channel separation.

    Parameters
    ----------
    time_series : np.ndarray
        The time series to be plotted. Shape must be (n_samples, n_channels).
    n_samples : int, optional
        The number of time points to be plotted.
    y_tick_values:, optional
        Labels for the channels to be placed on the y-axis.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    plot_kwargs : dict, optional
        Keyword arguments to be passed on to `ax.plot <https://matplotlib.org\
        /stable/api/_as_gen/matplotlib.axes.Axes.plot.html>`_. Defaults to
        :code:`{"lw": 0.7, "color": "tab:blue"}`.
    ax : plt.axes, optional
        The axis on which to plot the data. If not given, a new axis is created.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    time_series = np.asarray(time_series)
    n_samples = min(n_samples or np.inf, time_series.shape[0])
    n_channels = time_series.shape[1]

    # Validation
    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (12, 8)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}
    default_plot_kwargs = {"lw": 0.7, "color": "tab:blue"}
    plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

    # Calculate separation
    separation = (
        np.maximum(time_series[:n_samples].max(), time_series[:n_samples].min()) * 1.2,
    )
    gaps = np.arange(n_channels)[::-1] * separation

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot data
    ax.plot(time_series[:n_samples] + gaps[None, :], **plot_kwargs)

    ax.autoscale(tight=True)
    for spine in ax.spines.values():
        spine.set_visible(False)

    # Set x and y axis tick labels
    ax.set_xticks([])
    if y_tick_values is not None:
        ax.set_yticks(gaps)
        ax.set_yticklabels(y_tick_values)
    else:
        ax.set_yticks([])

    # Save figure
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_separate_time_series(
    *time_series,
    n_samples=None,
    sampling_frequency=None,
    fig_kwargs=None,
    plot_kwargs=None,
    filename=None,
):
    """Plot time series as separate subplots.

    Parameters
    ----------
    time_series : np.ndarray
        Time series to be plotted. Should be (n_samples, n_lines). Each line is
        its own subplot.
    sampling_frequency: float, optional
        Sampling frequency of the input data, enabling us to label the x-axis.
    n_samples : int, optional
        Number of samples to be shown on the x-axis.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    plot_kwargs : dict, optional
        Keyword arguments to be passed on to `ax.plot <https://matplotlib.org\
        /stable/api/_as_gen/matplotlib.axes.Axes.plot.html>`_. Defaults to
        :code:`{"lw": 0.7}`.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    time_series = np.asarray(time_series)
    n_samples = n_samples or min([ts.shape[0] for ts in time_series])
    n_lines = time_series[0].shape[1]

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (20, 10), "sharex": "all"}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}
    default_plot_kwargs = {"lw": 0.7}
    plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

    if sampling_frequency is not None:
        time_vector = np.linspace(0, n_samples / sampling_frequency, n_samples)
    else:
        time_vector = np.linspace(0, n_samples, n_samples)

    # Create figure
    fig, axes = create_figure(n_lines, **fig_kwargs)
    if n_lines == 1:
        axes = [axes]

    # Plot each time series
    for group in time_series:
        for axis, line in zip(axes, group.T):
            axis.plot(time_vector, line[:n_samples], **plot_kwargs)
            axis.autoscale(axis="x", tight=True)

    # Label the x-axis
    if sampling_frequency is not None:
        axes[-1].set_xlabel("Time (s)")
    else:
        axes[-1].set_xlabel("Sample")

    # Save figure
    if filename is not None:
        save(fig, filename, tight_layout=True)
    else:
        return fig, axes


def plot_epoched_time_series(
    data,
    time_index,
    sampling_frequency=None,
    pre=125,
    post=1000,
    baseline_correct=False,
    legend=True,
    legend_loc=1,
    title=None,
    plot_kwargs=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot continuous data, epoched and meaned over epochs.

    Parameters
    ----------
    data : np.ndarray
        Data to be epoched. Shape must be (n_samples, n_channels).
    time_index : np.ndarray
        The integer indices of the start of each epoch.
    sampling_frequency : float, optional
        The sampling frequency of the data in Hz.
    pre : int, optional
        The integer number of samples to include before the trigger.
    post : int, optional
        The integer number of samples to include after the trigger.
    baseline_correct : bool, optional
        Should we subtract the mean value pre-trigger.
    legend : bool, optional
        Should a legend be created.
    legend_loc : int, optional
        Location of the legend.
    title : str, optional
        Title of the figure.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    plot_kwargs : dict, optional
        Keyword arguments to be passed on to `ax.plot <https://matplotlib.org\
        /stable/api/_as_gen/matplotlib.axes.Axes.plot.html>`_.
    ax : plt.axes, optional
        The axis on which to plot the data. If not given, a new axis is created.
    filename : str, optional
        Output_filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    from osl_dynamics.data.task import epoch_mean

    epoched_1 = epoch_mean(data, time_index, pre, post)

    x_label = "Sample"
    time_index = np.arange(-pre, post)
    if sampling_frequency:
        time_index = time_index / sampling_frequency
        x_label = "Time (s)"

    # Validation
    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (16, 3)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Baseline correct
    if baseline_correct:
        epoched_1 -= np.mean(epoched_1[:pre], axis=0, keepdims=True)

    # Plot data
    for i, s in enumerate(epoched_1.T):
        ax.plot(time_index, s, label=i, **plot_kwargs)
    ax.axvline(0, c="k")
    ax.autoscale(axis="x", tight=True)

    # Set title and axis labels
    if title is not None:
        ax.set_title(title)
    ax.set_xlabel(x_label)

    # Add a legend
    if legend:
        ax.legend(loc=legend_loc)

    # Save the figure if a filename has been passed
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_matrices(
    matrix,
    group_color_scale=True,
    titles=None,
    main_title=None,
    cmap="viridis",
    nan_color="white",
    log_norm=False,
    filename=None,
):
    """Plot a collection of matrices.

    Given an iterable of matrices, plot each matrix in its own axis. The axes
    are arranged as close to a square (:code:`N x N` axis grid) as possible.

    Parameters
    ----------
    matrix: list of np.ndarray
        The matrices to plot.
    group_color_scale: bool, optional
        If True, all matrices will have the same colormap scale, where we use
        the minimum and maximum across all matrices as the scale.
    titles: list of str, optional
        Titles to give to each matrix axis.
    main_title: str, optional
        Main title to be placed at the top of the plot.
    cmap: str, optional
        Matplotlib colormap.
    nan_color: str, optional
        Matplotlib color to use for :code:`NaN` values.
    log_norm: bool, optional
        Should we show the elements on a log scale?
    filename: str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`filename=None`.
    """
    matrix = np.array(matrix)
    if matrix.ndim == 2:
        matrix = matrix[None, :]
    if matrix.ndim != 3:
        raise ValueError("Must be a 3D array.")
    short, long, empty = rough_square_axes(len(matrix))
    fig, axes = plt.subplots(ncols=short, nrows=long, squeeze=False)

    if titles is None:
        titles = [""] * len(matrix)

    cmap = matplotlib.cm.get_cmap(cmap).copy()
    cmap.set_bad(color=nan_color)

    for grid, axis, title in zip_longest(matrix, axes.ravel(), titles):
        if grid is None:
            axis.remove()
            continue
        if group_color_scale:
            v_min = np.nanmin(matrix)
            v_max = np.nanmax(matrix)
            if log_norm:
                im = axis.matshow(
                    grid,
                    cmap=cmap,
                    norm=matplotlib.colors.LogNorm(vmin=v_min, vmax=v_max),
                )
            else:
                im = axis.matshow(grid, vmin=v_min, vmax=v_max, cmap=cmap)
        else:
            if log_norm:
                im = axis.matshow(
                    grid,
                    cmap=cmap,
                    norm=matplotlib.colors.LogNorm(),
                )
            else:
                im = axis.matshow(grid, cmap=cmap)
        axis.set_title(title)
        if grid.shape[0] > 30:
            # Don't label the ticks if there's too many
            axis.set_xticklabels([])
            axis.set_yticklabels([])

    if group_color_scale:
        fig.subplots_adjust(right=0.8)
        color_bar_axis = fig.add_axes([0.85, 0.15, 0.05, 0.7])
        fig.colorbar(im, cax=color_bar_axis)
    else:
        for axis in fig.axes:
            pl = axis.get_images()[0]
            divider = make_axes_locatable(axis)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            plt.colorbar(pl, cax=cax)
        plt.tight_layout()

    fig.suptitle(main_title)

    if filename is not None:
        save(fig, filename)
    else:
        return fig, axes


def plot_connections(
    weights,
    labels=None,
    ax=None,
    cmap="hot",
    text_color=None,
    filename=None,
):
    """Create a chord diagram representing the values of a matrix.

    For a matrix of weights, create a chord diagram where the color of the line
    connecting two nodes represents the value indexed by the position of the
    nodes in the lower triangle of the matrix.

    This is useful for showing things like co-activation between sensors/parcels
    or relations between nodes in a network.

    Parameters
    ----------
    weights : np.ndarray
        An :code:`NxN` matrix of weights.
    labels : list of str, optional
        A name for each node in the weights matrix (e.g. parcel names)
    ax : plt.axes, optional
        A matplotlib axis on which to plot.
    cmap : str, optional
        A string corresponding to a matplotlib colormap.
    text_color : str, optional
        A string corresponding to a matplotlib color.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    weights = np.abs(weights)
    x, y = np.diag_indices_from(weights)
    weights[x, y] = 0
    weights /= weights.max()

    inner = 0.9
    outer = 1.0

    cmap = matplotlib.cm.get_cmap(cmap)
    norm = matplotlib.colors.Normalize(vmin=0, vmax=1)

    highest_color = cmap(norm(1)) if text_color is None else text_color
    zero_color = cmap(norm(0))

    text_color = {
        "text.color": highest_color,
        "axes.labelcolor": highest_color,
        "xtick.color": highest_color,
        "ytick.color": highest_color,
    }

    angle = np.radians(360 / weights.shape[0])
    pad = np.radians(0.5)

    starts = np.arange(0, 2 * np.pi, angle)
    lefts = starts + pad
    rights = starts + angle - pad
    centers = 0.5 * (lefts + rights)

    with matplotlib.rc_context(text_color):
        if not ax:
            fig = plt.figure(figsize=(10, 10))
            ax = fig.add_subplot(111, projection="polar")

        for left, right in zip(lefts, rights):
            verts = [
                (left, inner),
                (left, outer),
                (right, outer),
                (right, inner),
                (0.0, 0.0),
            ]

            codes = [
                Path.MOVETO,
                Path.LINETO,
                Path.LINETO,
                Path.LINETO,
                Path.CLOSEPOLY,
            ]

            path = Path(verts, codes)

            patch = patches.PathPatch(path, facecolor="orange", lw=1)
            ax.add_patch(patch)
        ax.set_yticks([])
        ax.grid(False)

        bezier_codes = [
            Path.MOVETO,
            Path.CURVE4,
            Path.CURVE4,
            Path.CURVE4,
        ]

        rebound = 0.5

        for i, j in zip(*np.tril_indices_from(weights)):
            center_1 = centers[i]
            center_2 = centers[j]

            verts = [
                (center_1, inner),
                (center_1, rebound),
                (center_2, rebound),
                (center_2, inner),
            ]

            path = Path(verts, bezier_codes)

            patch = patches.PathPatch(
                path,
                facecolor="none",
                lw=2,
                edgecolor=cmap(weights[i, j]),
                alpha=weights[i, j] ** 2,
            )
            ax.add_patch(patch)

        ax.set_xticks([])

        ax.set_facecolor(zero_color)
        fig.patch.set_facecolor(zero_color)

        if labels is None:
            labels = [""] * len(centers)
        for center, label in zip(centers, labels):
            rotation = np.degrees(center)

            if 0 <= rotation < 90:
                horizontal_alignment = "left"
                vertical_alignment = "bottom"
            elif 90 <= rotation < 180:
                horizontal_alignment = "right"
                vertical_alignment = "bottom"
            elif 180 <= rotation < 270:
                horizontal_alignment = "right"
                vertical_alignment = "top"
            else:
                horizontal_alignment = "left"
                vertical_alignment = "top"

            if 90 <= rotation < 270:
                rotation += 180

            ax.annotate(
                label,
                (center, outer + 0.05),
                rotation=rotation,
                horizontalalignment=horizontal_alignment,
                verticalalignment=vertical_alignment,
            )

        ax.autoscale_view()
        plt.setp(ax.spines.values(), visible=False)

    if filename is not None:
        save(fig, filename, tight_layout=True)
    else:
        return fig, ax


def topoplot(
    layout,
    data,
    channel_names=None,
    plot_boxes=False,
    show_deleted_sensors=False,
    show_names=False,
    title=None,
    colorbar=True,
    axis=None,
    cmap="cold_hot",
    n_contours=10,
    filename=None,
):
    """Make a contour plot in sensor space.

    Create a contour plot by interpolating a field from a set of values
    provided for each sensor location in an MEG layout. Within the context of
    DyNeMo this is likely to be an array of (all positive) values taken from
    the diagonal of a covariance matrix, but one can also plot any sensor
    level M/EEG data.

    Parameters
    ----------
    layout : str
        The name of an MEG layout (matching one from FieldTrip).
    data : np.ndarray
        The value of the field at each sensor.
    channel_names : list of str, optional
        A list of channel names which are present in the data
        (removes missing channels).
    plot_boxes : bool, optional
        Show boxes representing the height and width of sensors.
    show_deleted_sensors : bool, optional
        Show sensors missing from :code:`channel_names` in red.
    show_names : bool, optional
        Show the names of channels (can get very cluttered).
    title : str, optional
        A title for the figure.
    colorbar : bool, optional
        Show a colorbar for the field.
    axis : plt.axis, optional
        matplotlib axis to plot on.
    cmap : str, optional
        matplotlib colourmap.
    n_contours : int, optional
        number of field isolines to show on the plot.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    topology = Topology(layout)

    if channel_names is not None:
        topology.keep_channels(channel_names)

    fig = topology.plot_data(
        data,
        plot_boxes=plot_boxes,
        show_deleted_sensors=show_deleted_sensors,
        show_names=show_names,
        title=title,
        colorbar=colorbar,
        axis=axis,
        cmap=cmap,
        n_contours=n_contours,
    )

    if filename is not None:
        save(fig, filename, tight_layout=True)
    else:
        return fig


def plot_brain_surface(
    values,
    mask_file,
    parcellation_file,
    title=None,
    cmap="cold_hot",
    colorbar=True,
    symmetric_cbar=True,
    cbar_tick_format="%.2g",
    cbar_label=None,
    vmin=None,
    vmax=None,
    hemispheres=None,
    views=None,
    bg_on_data=False,
    threshold=None,
    filename=None,
    show_plot=None,
):
    """Plot a 2D heat map on the surface of the brain.

    Parameters
    ----------
    values : np.ndarray
        Data to plot. Must be of shape (n_parcels,).
    mask_file : str
        Mask file for the brain. See osl_dynamics.files.mask.
    parcellation_file : str
        Parcellation file. See osl_dynamics.files.parcellation.
    title : str, optional
        Title for the plot.
    cmap : str, optional
        Matplotlib colormap.
    colorbar : bool, optional
        Should we plot a colorbar?
    symmetric_cbar : bool, optional
        Should we have a symmetric color bar?
    cbar_tick_format : str, optional
        Formatting for the color bar tick labels.
        Example use: :code:`cbar_tick_format='%.2f'`.
    cbar_label : str, optional
        Label for the color bar.
    vmin : float, optional
        Minimum value for the color bar.
        May be overriden if :code:`symmetric_cbar=True`.
    vmax : float, optional
        Maximum value for the color bar.
        May be overriden if :code:`symmetric_cbar=True`.
    hemispheres : list, optional.
        :code:`['left', 'right']` or :code:`['left']` or :code:`['right']`.
        Defaults to :code:`['left', 'right']`.
    views : list, optional
        The list can contain :code:`'lateral'` or :code:`'medial'`.
        Defaults to :code:`['lateral']`, which will show one row with
        lateral views. :code:`['lateral', 'medial']` will show two
        rows with the lateral view ontop and medial view below.
    bg_on_data : bool, optional
        If True, the sulcal depth is jointly visible with surface data.
        Otherwise, the background image will only be visible where there
        is no surface data (either because the surface data contains nans
        or because is was thresholded).
    threshold : float, optional
        Threshold values to display. Defaults to no thresholding.
    filename : str, optional
        Output filename. Extension can be :code:`png/svg/pdf`.
        If None is passed then the image is shown on screen and the
        Matplotlib objects are returned.
    show_plot : bool, optional
        Should we show the plot? If :code:`filename` is True, defaults
        to False, otherwise False.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`filename` is None.
    ax : plt.axes
        Matplotlib axis object. Only returned if :code:`filename` is None.
    """
    if vmin is None:
        vmin = np.min(values)
    if vmax is None:
        vmax = np.max(values)
    if symmetric_cbar:
        vmax = np.max([vmax, -vmin])
        vmin = -vmax
    if hemispheres is None:
        hemispheres = ["left", "right"]
    if views is None:
        views = ["lateral"]

    if filename is not None:
        allowed_extensions = [".png", ".svg", ".pdf"]
        if not any([ext in filename for ext in allowed_extensions]):
            raise ValueError(
                "filename must have one of following extensions: "
                + f"{' '.join(allowed_extensions)}."
            )

    if filename is None:
        show_plot = True
    else:
        show_plot = False

    # Find files
    mask_file = files.check_exists(mask_file, files.mask.directory)
    parcellation_file = files.check_exists(
        parcellation_file, files.parcellation.directory
    )

    # Convert from parcel values to voxel values
    values = parcel_vector_to_voxel_grid(mask_file, parcellation_file, values)

    # Create image to plot
    mask = nib.load(mask_file)
    nii = nib.Nifti1Image(values, mask.affine, mask.header)

    # Plot
    fig, ax = plot_img_on_surf(
        nii,
        output_file=None,
        colorbar=False,
        cmap=cmap,
        symmetric_cbar=symmetric_cbar,
        vmin=vmin,
        vmax=vmax,
        hemispheres=hemispheres,
        views=views,
        bg_on_data=bg_on_data,
        threshold=threshold,
    )

    if views == ["lateral"]:
        # Plotting 2 views

        # Title
        fig.suptitle(title, fontsize=26, y=0.97)

        if colorbar:
            # Positioning and size of the colour bar
            # [left, bottom, width, height]
            cbar_ax = fig.add_axes([0.25, 0.2, 0.5, 0.05])

            # Colour bar
            sm = plt.cm.ScalarMappable(
                cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax)
            )
            cbar = fig.colorbar(
                sm,
                cax=cbar_ax,
                orientation="horizontal",
                format=cbar_tick_format,
            )
            cbar.ax.tick_params(labelsize=24)
            cbar.set_label(cbar_label, fontsize=24)
    else:
        # Plotting 4 views

        # Title
        fig.suptitle(title, fontsize=18, y=0.98)

        if colorbar:
            # Positioning and size of the colour bar
            # [left, bottom, width, height]
            cbar_ax = fig.add_axes([0.3, 0.1, 0.4, 0.04])

            # Colour bar
            sm = plt.cm.ScalarMappable(
                cmap=cmap, norm=plt.Normalize(vmin=vmin, vmax=vmax)
            )
            cbar = fig.colorbar(
                sm,
                cax=cbar_ax,
                orientation="horizontal",
                format=cbar_tick_format,
            )
            cbar.ax.tick_params(labelsize=16)
            cbar.set_label(cbar_label, fontsize=16)

    # Save or return
    if filename is not None:
        fig.savefig(filename)
        if not show_plot:
            plt.close(fig)
    else:
        return fig, ax


def plot_alpha(
    *alpha,
    n_samples=None,
    cmap="tab10",
    sampling_frequency=None,
    y_labels=None,
    title=None,
    plot_kwargs=None,
    fig_kwargs=None,
    filename=None,
    axes=None,
):
    """Plot alpha.

    Parameters
    ----------
    alpha : np.ndarray
        A collection of alphas passed as separate arguments.
    n_samples: int, optional
        Number of time points to be plotted.
    cmap : str, optional
        A matplotlib colormap string.
    sampling_frequency : float, optional
        The sampling frequency of the data in Hz.
    y_labels : str, optional
        Labels for the y-axis of each alpha time series.
    title : str, optional
        Title for the plot.
    plot_kwargs : dict, optional
        Any parameters to be passed to `plt.stackplot <https://matplotlib.org\
        /stable/api/_as_gen/matplotlib.pyplot.stackplot.html>`_.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    filename : str, optional
        Output filename.
    axes : list of plt.axes, optional
        A list of matplotlib axes to plot on. If :code:`None`, a new
        figure is created.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    n_alphas = len(alpha)
    if isinstance(axes, plt.Axes):
        axes = [axes]
    if axes is not None and len(axes) != n_alphas:
        raise ValueError("Number of axes must match number of alphas.")

    n_modes = max(a.shape[1] for a in alpha)
    n_samples = min(n_samples or np.inf, alpha[0].shape[0])
    if cmap in [
        "Pastel1",
        "Pastel2",
        "Paired",
        "Accent",
        "Dark2",
        "Set1",
        "Set2",
        "Set3",
        "tab10",
        "tab20",
        "tab20b",
        "tab20c",
    ]:
        cmap = plt.cm.get_cmap(name=cmap)
    else:
        cmap = plt.cm.get_cmap(name=cmap, lut=n_modes)
    colors = cmap.colors

    # Validation
    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = dict(
        figsize=(12, 2.5 * n_alphas), sharex="all", facecolor="white"
    )
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}
    default_plot_kwargs = dict(colors=colors)
    plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)

    if y_labels is None:
        y_labels = [None] * n_alphas
    elif isinstance(y_labels, str):
        y_labels = [y_labels] * n_alphas
    elif len(y_labels) != n_alphas:
        raise ValueError("Incorrect number of y_labels passed.")

    # Create figure if axes not passed
    if axes is None:
        fig, axes = create_figure(n_alphas, **fig_kwargs)
    else:
        fig = axes[0].get_figure()

    if isinstance(axes, plt.Axes):
        axes = [axes]

    # Plot data
    for a, ax, y_label in zip(alpha, axes, y_labels):
        time_vector = (
            np.arange(n_samples) / sampling_frequency
            if sampling_frequency
            else range(n_samples)
        )
        ax.stackplot(time_vector, a[:n_samples].T, **plot_kwargs)
        ax.autoscale(tight=True)
        ax.set_ylabel(y_label)

    # Set axis label and title
    axes[-1].set_xlabel("Time (s)" if sampling_frequency else "Sample")
    axes[0].set_title(title)

    # Fix layout
    plt.tight_layout()

    # Add a colour bar
    norm = matplotlib.colors.BoundaryNorm(
        boundaries=range(n_modes + 1), ncolors=n_modes
    )
    mappable = plt.cm.ScalarMappable(norm=norm, cmap=cmap)
    fig.subplots_adjust(right=0.94)
    cb_ax = fig.add_axes([0.95, 0.15, 0.025, 0.7])
    cb = fig.colorbar(mappable, cax=cb_ax, ticks=np.arange(0.5, n_modes, 1))
    cb.ax.set_yticklabels(range(1, n_modes + 1))

    # Save to file if a filename as been passed
    if filename is not None:
        save(fig, filename)
    else:
        return fig, axes


def plot_state_lifetimes(
    state_time_course,
    bins="auto",
    density=False,
    match_scale_x=False,
    match_scale_y=False,
    x_range=None,
    x_label=None,
    y_label=None,
    plot_kwargs=None,
    fig_kwargs=None,
    filename=None,
):
    """Create a histogram of state lifetimes.

    For a state time course, create a histogram for each state with the
    distribution of the lengths of time for which it is active.

    Parameters
    ----------
    state_time_course : np.ndarray
        Mode time course to analyse.
    bins : int, optional
        Number of bins for the histograms.
    density : bool, optional
        If :code:`True`, plot the probability density of the state activation
        lengths. If :code:`False`, raw number.
    match_scale_x : bool, optional
        If True, all histograms will share the same x-axis scale.
    match_scale_y : bool, optional
        If True, all histograms will share the same y-axis scale.
    x_range : list, optional
        The limits on the values presented on the x-axis.
    x_label : str, optional
        x-axis label.
    y_label : str, optional
        y-axis label.
    plot_kwargs : dict, optional
        Keyword arguments to pass to `ax.hist <https://matplotlib.org/stable\
        /api/_as_gen/matplotlib.axes.Axes.hist.html>`_.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    from osl_dynamics.analysis import modes

    n_plots = state_time_course.shape[1]
    short, long, empty = rough_square_axes(n_plots)
    colors = get_colors(n_plots)

    # Validation
    if state_time_course.ndim == 1:
        state_time_course = get_one_hot(state_time_course)
    if state_time_course.ndim != 2:
        raise ValueError("state_time_course must be a 2D array.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (long * 2.5, short * 2.5)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    if plot_kwargs is None:
        plot_kwargs = {}

    # Calculate state lifetimes
    channel_lifetimes = modes.lifetimes(state_time_course)

    # Create figure
    fig, axes = create_figure(short, long, **fig_kwargs)

    # Plot data
    largest_bar = 0
    furthest_value = 0
    for channel, axis, color in zip_longest(channel_lifetimes, axes.ravel(), colors):
        if channel is None:
            axis.remove()
            continue
        if not len(channel):
            axis.text(
                0.5,
                0.5,
                "No\nactivation",
                horizontalalignment="center",
                verticalalignment="center",
                transform=axis.transAxes,
                fontsize=20,
            )
            axis.set_xticks([])
            axis.set_yticks([])
            continue
        hist = axis.hist(
            channel, density=density, bins=bins, color=color, **plot_kwargs
        )
        largest_bar = max(hist[0].max(), largest_bar)
        furthest_value = max(hist[1].max(), furthest_value)
        t = axis.text(
            0.95,
            0.95,
            f"{np.sum(channel) / len(state_time_course) * 100:.2f}%",
            fontsize=10,
            horizontalalignment="right",
            verticalalignment="top",
            transform=axis.transAxes,
        )
        axis.xaxis.set_tick_params(labelbottom=True, labelleft=True)
        t.set_bbox({"facecolor": "white", "alpha": 0.7, "boxstyle": "round"})

    # Set axis range and labels
    for axis in axes.ravel():
        if match_scale_x:
            axis.set_xlim(0, furthest_value * 1.1)
        if match_scale_y:
            axis.set_ylim(0, largest_bar * 1.1)
        if x_range is not None:
            if len(x_range) != 2:
                raise ValueError("x_range must be [x_min, x_max].")
            axis.set_xlim(x_range[0], x_range[1])
        axis.set_xlabel(x_label)
        axis.set_ylabel(y_label)

    # Save file is a filename has been passed
    if filename is not None:
        save(fig, filename, tight_layout=True)
    else:
        return fig, axes


def plot_psd_topo(
    f,
    psd,
    only_show=None,
    parcellation_file=None,
    topomap_pos=None,
    fig_kwargs=None,
    filename=None,
):
    """PLot PSDs for parcels and a topomap.

    Parameters
    ----------
    f : np.ndarray
        Frequency axis. Shape must be (n_freq,).
    psd : np.ndarray
        PSD for each parcel. Shape must be (n_parcels, n_freq).
    only_show : list, optional
        Indices for parcels to include in the plot. Defaults to all parcels.
    parcellation_file : str, optional
        Path to parcellation file.
    topomap_pos : list, optional
        Positioning and size of the topomap: :code:`[x0, y0, width, height]`.
        :code:`x0`, :code:`y0`, :code:`width`, :code:`height` should be floats
        between 0 and 1. Defaults to :code:`[0.45, 0.55, 0.5, 0.55]` to place
        the topomap on the top right. This is not used if
        :code:`parcellation_file=None`.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    if topomap_pos is None:
        topomap_pos = [0.45, 0.55, 0.5, 0.55]

    if fig_kwargs is None:
        fig_kwargs = {}

    if parcellation_file is not None:
        # Get the center of each parcel
        parcellation = Parcellation(parcellation_file)
        roi_centers = parcellation.roi_centers()

        # Re-order to use colour to indicate anterior->posterior location
        order = np.argsort(roi_centers[:, 1])
        roi_centers = roi_centers[order]
        psd = np.copy(psd)[order]

    n_parcels = psd.shape[0]

    # Which parcels should we plot the PSD for?
    if only_show is None:
        only_show = np.arange(n_parcels)

    # Plot PSDs
    fig, ax = create_figure(**fig_kwargs)
    cmap = plt.cm.viridis_r
    for i in reversed(range(n_parcels)):
        if i in only_show:
            ax.plot(f, psd[i], c=cmap(i / n_parcels))
    ax.set_xlabel("Frequency (Hz)")
    ax.set_ylabel("PSD (a.u.)")
    ax.set_xlim(f[0], f[-1])
    plt.tight_layout()

    if parcellation_file is not None:
        # Plot parcel topomap
        inside_ax = ax.inset_axes(topomap_pos)
        plot_markers(
            np.arange(parcellation.n_parcels),
            roi_centers,
            node_size=12,
            colorbar=False,
            axes=inside_ax,
        )

    # Save
    if filename is not None:
        save(fig, filename)
    else:
        return fig, ax


def plot_hmm_summary_stats(
    fo,
    lt,
    intv,
    sr,
    filename=None,
    cmap="tab10",
    fig_kwargs=None,
    sns_kwargs=None,
):
    """Plot summary statistics (FO, LT, INTV, SR).

    Parameters
    ----------
    fo : np.ndarray
        Fractional occupancies. Shape must be (n_subjects, n_states).
    lt : np.ndarray
        Mean lifetimes in seconds. Shape must be (n_subjects, n_states).
    intv : np.ndarray
        Mean intervals in seconds. Shape must be (n_subjects, n_states).
    sr : np.ndarray
        Switching rates in Hz. Shape must be (n_subjects, n_states).
    filename : str, optional
        Output filename.
    cmap : str, optional
        Matplotlib colormap.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    sns_kwargs : dict, optional
        Arguments to pass to :code:`sns.violinplot()`.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    if fig_kwargs is None:
        fig_kwargs = {}

    if sns_kwargs is None:
        sns_kwargs = {}

    n_states = fo.shape[1]
    x = range(1, n_states + 1)

    sns_kwargs.update(
        {
            "inner": "quart",
            "cut": 0,
            "palette": cmap,
        }
    )

    fig, ax = create_figure(nrows=1, ncols=4, figsize=(15, 3))
    plot_violin(
        fo.T,
        x=x,
        x_label="State",
        y_label="Fractional Occupancy",
        ax=ax[0],
        sns_kwargs=sns_kwargs,
    )
    plot_violin(
        lt.T,
        x=x,
        x_label="State",
        y_label="Mean Lifetime (s)",
        ax=ax[1],
        sns_kwargs=sns_kwargs,
    )
    plot_violin(
        intv.T,
        x=x,
        x_label="State",
        y_label="Mean Interval (s)",
        ax=ax[2],
        sns_kwargs=sns_kwargs,
    )
    plot_violin(
        sr.T,
        x=x,
        x_label="State",
        y_label="Switching rate (Hz)",
        ax=ax[3],
        sns_kwargs=sns_kwargs,
    )
    if filename is None:
        fig.tight_layout()
        return fig, ax
    save(fig, filename=filename, tight_layout=True)


def plot_summary_stats_group_diff(
    name,
    summary_stats,
    pvalues,
    assignments,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot summary statistics for two groups as violin plots.

    Parameters
    ----------
    name : str
        Name of the summary statistic.
    summary_stats : np.ndarray
        Summary statistics. Shape is (n_sessions, n_states).
    pvalues : np.ndarray
        p-values for each summary statistic difference. Shape is (n_states,).
    assignments : np.ndarray
        Array of 1s and 2s indicating group assignment. Shape is (n_sessions,).
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    # Validation
    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    # Create a pandas DataFrame to hold the summary stats
    ss_dict = {name: [], "State": [], "Group": []}
    n_sessions, n_states = summary_stats.shape
    for array in range(n_sessions):
        for state in range(n_states):
            ss_dict[name].append(summary_stats[array, state])
            ss_dict["State"].append(state + 1)
            ss_dict["Group"].append(assignments[array])
    ss_df = pd.DataFrame(ss_dict)

    # Create figure
    if fig_kwargs is None:
        fig_kwargs = {}
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot a half violin for each group
    sns.violinplot(
        data=ss_df,
        x="State",
        y=name,
        hue="Group",
        split=True,
        inner=None,
        ax=ax,
    )

    # Add a star above the violin to indicate significance
    scatter_kwargs = {"c": "black", "s": 32, "marker": "*"}
    for i in range(n_states):
        if pvalues[i] < 0.01:
            ax.scatter(
                i - 0.075,
                summary_stats[:, i].max() * 1.6,
                **scatter_kwargs,
            )
            ax.scatter(
                i + 0.075,
                summary_stats[:, i].max() * 1.6,
                **scatter_kwargs,
            )
        elif pvalues[i] < 0.05:
            ax.scatter(i, summary_stats[:, i].max() * 1.6, **scatter_kwargs)

    # Save figure
    if filename is not None:
        _logger.info(f"Saving {filename}")
        plt.savefig(filename)
        plt.close()
    elif create_fig:
        return fig, ax


def plot_evoked_response(
    t,
    epochs,
    pvalues,
    significance_level=0.05,
    offset_between_bars=0.01,
    labels=None,
    legend_loc=1,
    x_label=None,
    y_label=None,
    title=None,
    fig_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot an evoked responses with significant time points highlighted.

    Parameters
    ----------
    t : np.ndarray
        Time axis. Shape must be (n_samples,).
    epochs : np.ndarray
        Evoked responses. Shape must be (n_samples, n_channels).
    pvalues : np.ndarray
        p-value for each evoked response. This can be calculated with
        `osl_dynamics.analysis.statistics.evoked_response_max_stat_perm 
        <https://osl-dynamics.readthedocs.io/en/latest/autoapi/osl_dynamics\
        /analysis/statistics/index.html#osl_dynamics.analysis.statistics\
        .evoked_response_max_stat_perm>`_.
    significance_level : float, optional
        Value to threshold the p-values with to consider significant.
        By default :code:`pvalues < 0.05` are significant.
    offset_between_bars : float, optional
        Vertical offset between bars that highlight significance.
    labels : list, optional
        Label for each evoked response time series.
    legend_loc : int, optional
        Position of the legend.
    x_label : str, optional
        Label for x-axis.
    y_label : str, optional
        Label for y-axis.
    title : str, optional
        Figure title.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """

    # Validation
    if labels is not None:
        if isinstance(labels, str):
            labels = [labels]
        else:
            if len(labels) != epochs.shape[1]:
                raise ValueError("Incorrect number of lines or labels passed.")
        add_legend = True
    else:
        labels = [None] * epochs.shape[1]
        add_legend = False

    if ax is not None:
        if filename is not None:
            raise ValueError(
                "Please use plotting.save() to save the figure instead of the "
                + "filename argument."
            )
        if isinstance(ax, np.ndarray):
            raise ValueError("Only pass one axis.")

    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (7, 4)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)

    # Get significant time points for each channel
    significant = pvalues < significance_level

    # Create figure
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    for i, e, l, s in zip(
        range(epochs.shape[1]),
        epochs.T,
        labels,
        significant.T,
    ):
        # Plot evoked response
        p = ax.plot(t, e, label=l)

        # Highlight significant time points
        sig_times = t[s]
        if len(sig_times) > 0:
            y = 1.1 * np.max(epochs) + i * offset_between_bars
            dt = (t[1] - t[0]) / 2
            for st in sig_times:
                ax.plot(
                    (st - dt, st + dt),
                    (y, y),
                    color=p[0].get_color(),
                    linewidth=3,
                )

    # Add a dashed line at time = 0
    ax.axvline(0, linestyle="--", color="black")

    # Set title, axis labels and range
    ax.set_title(title)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.set_xlim(t[0], t[-1])

    # Add a legend
    if add_legend:
        ax.legend(loc=legend_loc)

    # Save figure
    if filename is not None:
        save(fig, filename, tight_layout=True)
    elif create_fig:
        return fig, ax


def plot_wavelet(
    data,
    sampling_frequency,
    w=5,
    standardize=True,
    time_range=None,
    frequency_range=None,
    title=None,
    add_colorbar=True,
    fig_kwargs=None,
    plot_kwargs=None,
    ax=None,
    filename=None,
):
    """Plot a wavelet transform.

    Parameters
    ----------
    data : np.ndarray
        1D time series data.
    sampling_frequency : float
        Sampling frequency in Hz.
    w : float, optional
        :code:`w` parameter to pass to `scipy.signal.morlet2 
        <https://docs.scipy.org/doc/scipy/reference/generated\
        /scipy.signal.morlet2.html>`_.
    standardize : bool, optional
        Should we standardize the data before calculating the wavelet?
    time_range : list, optional
        Start time and end time to plot in seconds.
        Default is the full time axis of the data.
    frequency_range : list of length 2, optional
        Start and end frequency to plot in Hz.
        Default is :code:`[1, sampling_frequency / 2]`.
    title : str, optional
        Figure title.
    add_colorbar : bool, optional
        If :code:`True` (default), space will be stolen from the figure to
        create a colorbar.
    fig_kwargs : dict, optional
        Arguments to pass to :code:`plt.subplots()`.
    plot_kwargs : dict, optional
        Keyword arguments to pass to `ax.pcolormesh 
        <https://matplotlib.org/stable/api/_as_gen/matplotlib.axes\
        .Axes.pcolormesh.html>`_. Defaults to :code:`{"cmap": "rainbow"}`.
    ax : plt.axes, optional
        Axis object to plot on.
    filename : str, optional
        Output filename.

    Returns
    -------
    fig : plt.figure
        Matplotlib figure object. Only returned if :code:`ax=None` and
        :code:`filename=None`.
    ax : plt.axes
        Matplotlib axis object(s). Only returned if :code:`ax=None` and
        :code:`filename=None`.
    """
    from osl_dynamics.analysis import spectral

    # Calculate wavelet transform
    t, f, wt = spectral.wavelet(
        data=data,
        sampling_frequency=sampling_frequency,
        w=w,
        standardize=standardize,
        time_range=time_range,
        frequency_range=frequency_range,
    )

    # Create figure
    if fig_kwargs is None:
        fig_kwargs = {}
    default_fig_kwargs = {"figsize": (12, 3)}
    fig_kwargs = override_dict_defaults(default_fig_kwargs, fig_kwargs)
    create_fig = ax is None
    if create_fig:
        fig, ax = create_figure(**fig_kwargs)

    # Plot
    if plot_kwargs is None:
        plot_kwargs = {}
    default_plot_kwargs = {"cmap": "rainbow"}
    plot_kwargs = override_dict_defaults(default_plot_kwargs, plot_kwargs)
    mappable = ax.pcolormesh(t, f, wt, **plot_kwargs)

    if add_colorbar:
        plt.subplots_adjust(bottom=0.2, right=0.8, top=0.9)
        cax = plt.axes([0.825, 0.1, 0.025, 0.8])
        plt.colorbar(mappable=mappable, cax=cax)

    # Set title and axis labels
    ax.set_title(title)
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Frequency (Hz)")

    # Save figure
    if filename is not None:
        save(fig, filename)
    elif create_fig:
        return fig, ax



"""Implementation of the Fisher kernel for prediction studies.

See the `HMM description <https://osl-dynamics.readthedocs.io/en/latest/models\
/hmm.html>`_ for further details.

See Also
--------
`Example script <https://github.com/OHBA-analysis/osl-dynamics/blob/main\
/examples/simulation/hmm_hmm-mvn_fisher-kernel.py>`_ applying the Fisher kernel
to simulated HMM data.
"""

import logging


import numpy as np
from tqdm.auto import trange

_logger = logging.getLogger("osl-dynamics")


class FisherKernel:
    """Class for computing the Fisher kernel matrix given a generative model.

    Parameters
    ----------
    model : osl-dynamics model
        Model. Currently only the :code:`HMM`, :code:`DyNeMo` and
        :code:`M-DyNeMo` are implemented.
    """

    def __init__(self, model):
        compatible_models = ["HMM", "DyNeMo", "M-DyNeMo"]
        if model.config.model_name not in compatible_models:
            raise NotImplementedError(
                f"{model.config.model_name} was not found."
                + f"Options are {compatible_models}."
            )

        self.model = model

    def get_features(self, dataset, batch_size=None):
        """Get the Fisher features.

        Parameters
        ----------
        dataset : osl_dynamics.data.Data
            Data.
        batch_size : int, optional
            Batch size. If :code:`None`, we use :code:`model.config.batch_size`.

        Returns
        -------
        features : np.ndarray
            Fisher kernel matrix. Shape is (n_sessions, n_features).
        """
        _logger.info("Getting Fisher features")

        n_sessions = dataset.n_sessions
        if batch_size is not None:
            self.model.config.batch_size = batch_size

        dataset = self.model.make_dataset(
            dataset,
            concatenate=False,
            shuffle=False,
        )

        # Initialise list to hold features for each session
        features = []
        for i in trange(n_sessions, desc="Getting features"):
            session_data = dataset[i]

            # Initialise dictionary for holding gradients
            d_model = dict()

            if self.model.config.model_name == "HMM":
                d_model["d_initial_distribution"] = []
                d_model["d_trans_prob"] = []

            trainable_variable_names = [
                var.name for var in self.model.trainable_weights
            ]
            # Get only variable in the generative model
            for name in trainable_variable_names:
                if (
                    "mod" in name
                    or "alpha" in name
                    or "gamma" in name
                    or "means" in name
                    or "covs" in name
                    or "stds" in name
                    or "fcs" in name
                ):
                    d_model[name] = []

            # Loop over data for each session
            for inputs in session_data:
                if self.model.config.model_name == "HMM":
                    x = inputs["data"]
                    gamma, xi = self.model.get_posterior(x)
                    d_initial_distribution, d_trans_prob = self._d_HMM(gamma, xi)
                    d_model["d_initial_distribution"].append(d_initial_distribution)
                    d_model["d_trans_prob"].append(d_trans_prob)
                    inputs = np.concatenate(
                        [x, gamma.reshape(x.shape[0], x.shape[1], -1)], axis=2
                    )

                gradients = self._get_tf_gradients(inputs)

                for name in d_model.keys():
                    if name == "d_initial_distribution" or name == "d_trans_prob":
                        continue
                    d_model[name].append(gradients[name])

            # Concatenate the flattened gradients
            session_features = np.concatenate(
                [np.sum(grad, axis=0).flatten() for grad in d_model.values()]
            )
            features.append(session_features)

        features = np.array(features)  # shape=(n_sessions, n_features)

        # Normalise the features to l2-norm of 1
        features_l2_norm = np.sqrt(np.sum(np.square(features), axis=-1, keepdims=True))
        features /= features_l2_norm
        return features

    def get_kernel_matrix(self, dataset, batch_size=None):
        """Get the Fisher kernel matrix.

        Parameters
        ----------
        dataset : osl_dynamics.data.Data
            Data.
        batch_size : int, optional
            Batch size. If :code:`None`, we use :code:`model.config.batch_size`.

        Returns
        -------
        kernel_matrix : np.ndarray
            Fisher kernel matrix. Shape is (n_sessions, n_sessions).
        """
        _logger.info("Getting Fisher kernel matrix")

        features = self.get_features(dataset, batch_size=batch_size)

        # Compute the kernel matrix with inner product
        kernel_matrix = features @ features.T

        return kernel_matrix

    def _d_HMM(self, gamma, xi):
        """Get the derivative of free energy with respect to
        transition probability, initial distribution of HMM.

        Parameters
        ----------
        gamma : np.ndarray
            Marginal posterior distribution of hidden states given the data.
            Shape is (batch_size*sequence_length, n_states).
        xi : np.ndarray
            Joint posterior distribution of hidden states given the data.
            Shape is (batch_size*sequence_length-1, n_states*n_states).

        Returns
        -------
        d_initial_distribution : np.ndarray
            Derivative of free energy with respect to the initial distribution.
            Shape is (n_states,).
        d_trans_prob : np.ndarray
            Derivative of free energy with respect to the transition
            probability. Shape is (n_states, n_states).
        """
        initial_distribution = self.model.state_probs_t0
        initial_distribution = np.maximum(initial_distribution, 1e-6)
        initial_distribution /= np.sum(initial_distribution)

        trans_prob = self.model.trans_prob
        n_states = trans_prob.shape[0]
        xi = np.reshape(xi, (xi.shape[0], n_states, n_states), order="F")

        trans_prob = np.maximum(trans_prob, 1e-6)
        trans_prob /= np.sum(trans_prob, axis=1, keepdims=True)

        d_initial_distribution = gamma[0] / initial_distribution
        d_trans_prob = np.sum(xi, axis=0) / trans_prob
        return d_initial_distribution, d_trans_prob

    def _get_tf_gradients(self, inputs):
        """Get the gradient with respect to means and covariances.

        Parameters
        ----------
        inputs : tf.data.Dataset
            Model inputs.

        Returns
        -------
        gradients : dict
            Gradients with respect to the trainable variables.
        """
        import tensorflow as tf  # avoid slow imports

        with tf.GradientTape() as tape:
            loss = self.model.model(inputs)
            trainable_weights = {var.name: var for var in self.model.trainable_weights}
            gradients = tape.gradient(loss, trainable_weights)

        return gradients



"""Functions to perform spectral analysis."""

import logging

import numpy as np
from scipy import signal
from sklearn.decomposition import non_negative_factorization
from pqdm.processes import pqdm
from tqdm.auto import trange

from osl_dynamics import array_ops
from osl_dynamics.analysis import regression
from osl_dynamics.utils.misc import nextpow2

_logger = logging.getLogger("osl-dynamics")


def wavelet(
    data,
    sampling_frequency,
    w=5,
    standardize=True,
    time_range=None,
    frequency_range=None,
    df=0.5,
):
    """Calculate a wavelet transform.

    This function uses a Morlet2 window to calculate a wavelet transform.

    Parameters
    ----------
    data : np.ndarray
        1D time series data. Shape must be (n_samples,).
    sampling_frequency : float
        Sampling frequency in Hz.
    w : float, optional
        Number of oscillations in the Gaussian envelope.
    standardize : bool, optional
        Should we standardize the data before calculating the wavelet?
    time_range : list, optional
        Start time and end time to plot in seconds.
        Default is the full time axis of the data.
    frequency_range : list of length 2, optional
        Start and end frequency to plot in Hz.
        Default is :code:`[1, sampling_frequency / 2]`.
    df : float, optional
        Frequency resolution in Hz.

    Returns
    -------
    t : np.ndarray
        1D numpy array for the time axis.
    f : np.ndarray
        1D numpy array for the frequency axis.
    wt : np.ndarray
        2D numpy array (frequency, time) containing the wavelet transform.
    """
    # Validation
    if np.array(data).ndim != 1:
        raise ValueError("data must be a 1D numpy array.")

    if time_range is None:
        time_range = [0, data.shape[0] / sampling_frequency]
    if time_range[0] is None:
        time_range[0] = 0
    if time_range[1] is None:
        time_range[1] = data.shape[0] / sampling_frequency

    if frequency_range is None:
        frequency_range = [1, sampling_frequency / 2]
    if frequency_range[0] is None:
        frequency_range[0] = 1
    if frequency_range[1] is None:
        frequency_range[1] = sampling_frequency / 2

    # Standardize the data
    if standardize:
        data = (data - np.mean(data)) / np.std(data)

    # Keep selected time points
    start_index = int(time_range[0] * sampling_frequency)
    end_index = int(time_range[1] * sampling_frequency)
    data = data[start_index:end_index]

    # Time axis (s)
    dt = 1 / sampling_frequency
    t = np.arange(time_range[0], time_range[1], dt)

    # Frequency axis (Hz)
    f = np.arange(frequency_range[0], frequency_range[1], df)

    # Calculate the width for each Morlet window based on the frequency
    widths = w * sampling_frequency / (2 * f * np.pi)

    def _morlet2(M, s):
        x = np.arange(0, M) - (M - 1.0) / 2
        x = x / s
        sinusoidal = np.exp(1j * w * x)
        gaussian = np.exp(-0.5 * x**2)
        normalization = np.pi ** (-0.25) * np.sqrt(1 / s)
        return sinusoidal * gaussian * normalization

    def _cwt(data, widths):
        output = np.empty((len(widths), len(data)), dtype=np.complex64)
        for i, width in enumerate(widths):
            M = np.min([10 * width, len(data)])
            wavelet_data = np.conj(_morlet2(M, width)[::-1])
            output[i] = np.convolve(data, wavelet_data, mode="same")
        return output

    # Calculate wavelet transform
    wt = _cwt(data, widths)
    wt = abs(wt)

    return t, f, wt


def welch_spectra(
    data,
    sampling_frequency,
    alpha=None,
    window_length=None,
    step_size=None,
    frequency_range=None,
    standardize=True,
    calc_coh=True,
    return_weights=False,
    keepdims=False,
    n_jobs=1,
):
    """Calculates spectra for inferred states using Welch's method.

    The scaling for the power spectra calculated by this function
    matches SciPy (:code:`scipy.signal.welch`).

    Parameters
    ----------
    data : np.ndarray or list
        Time series data. Must have shape (n_sessions, n_samples,
        n_channels) or (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    alpha : np.ndarray or list, optional
        Inferred state probability time course. Must have shape
        (n_sessions, n_samples, n_states) or (n_samples, n_states).
    window_length : int, optional
        Length of the data segment to use to calculate spectra.
        If None, we use :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length // 2` (50% overlap).
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectra?
    calc_coh : bool, optional
        Should we also return the coherence spectra?
    return_weights : bool, optional
        Should we return the weights for session-specific spectra?
        This is useful for calculating a group average.
    keepdims : bool, optional
        Should we enforce a (n_sessions, n_states, ...) array is returned
        for :code:`psd` and :code:`coh`? If :code:`False`, we remove any
        dimension of length 1.
    n_jobs : int, optional
        Number of parallel jobs.

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences. Shape is (n_freq,).
    psd : np.ndarray
        Power spectra for each session and state. Shape is (n_sessions,
        n_states, n_channels, n_freq). Any axis of length 1 is removed if
        :code:`keepdims=False`.
    coh : np.ndarray
        Coherences for each session and state. Shape is (n_sessions, n_states,
        n_channels, n_channels, n_freq). Any axis of length 1 is removed if
        :code:`keepdims=False`. Only returned is :code:`calc_coh=True`.
    w : np.ndarray
        Weighting for session-specific spectra. Only returned if
        :code:`return_weights=True`. Shape is (n_sessions,).
    """

    # Validation
    if alpha is not None:
        if (isinstance(data, list) != isinstance(alpha, list)) or (
            isinstance(data, np.ndarray) != isinstance(alpha, np.ndarray)
        ):
            raise ValueError(
                f"data is type {type(data)} and alpha is type "
                f"{type(alpha)}. They must both be lists or numpy arrays."
            )

        if isinstance(data, list):
            # Check data and state time course for the same number of arrays
            # has been passed
            if len(data) != len(alpha):
                raise ValueError(
                    "A different number of arrays has been passed for "
                    f"data and alpha: len(data)={len(data)}, "
                    f"len(alpha)={len(alpha)}."
                )

            # Check the number of samples in data and alpha
            for i in range(len(alpha)):
                if alpha[i].shape[0] != data[i].shape[0]:
                    raise ValueError(
                        "items in data and alpha must have the number of samples."
                    )

    elif isinstance(data, list):
        alpha = np.array([None] * len(data))

    if isinstance(data, np.ndarray):
        if alpha is not None and alpha.shape[0] != data.shape[0]:
            raise ValueError("data and alpha must have the number of samples.")

        if data.ndim == 2:
            data = [data]
            alpha = [alpha]

    if window_length is None:
        window_length = 2 * sampling_frequency
    window_length = int(window_length)

    if step_size is None:
        step_size = window_length // 2
    step_size = int(step_size)

    if frequency_range is None:
        frequency_range = [0, sampling_frequency / 2]

    # Create keyword arguments to pass to the main function
    # that calculates spectra
    kwargs = []
    for d, a in zip(data, alpha):
        kwargs.append(
            {
                "data": d,
                "sampling_frequency": sampling_frequency,
                "alpha": a,
                "window_length": window_length,
                "step_size": step_size,
                "frequency_range": frequency_range,
                "standardize": standardize,
                "calc_coh": calc_coh,
                "keepdims": True,
            }
        )

    # Main calculation of spectra
    if len(data) == 1:
        # We only have one array so we don't need to parallelise
        # the calculation
        _logger.info("Calculating spectra")
        results = [_welch(**kwargs[0])]

    elif n_jobs == 1:
        # We have multiple arrays but we're running in serial
        results = []
        for i in trange(len(data), desc="Calculating spectra"):
            results.append(_welch(**kwargs[i]))

    else:
        # Calculate spectra in parallel
        _logger.info("Calculating spectra")
        results = pqdm(kwargs, _welch, n_jobs=n_jobs, argument_type="kwargs")

    # Unpack the results
    if calc_coh:
        psd = []
        coh = []
        for result in results:
            f, p, c = result
            psd.append(p)
            coh.append(c)
    else:
        psd = []
        for result in results:
            f, p = result
            psd.append(p)

    if not keepdims:
        # Remove any axes that are of length 1
        psd = np.squeeze(psd)
        if calc_coh:
            coh = np.squeeze(coh)

    # Weights for calculating a group-average spectrum
    n_samples = [d.shape[0] for d in data]
    w = np.array(n_samples) / np.sum(n_samples)

    if calc_coh:
        if return_weights:
            return f, psd, coh, w
        else:
            return f, psd, coh
    else:
        if return_weights:
            return f, psd, w
        else:
            return f, psd


def multitaper_spectra(
    data,
    sampling_frequency,
    alpha=None,
    window_length=None,
    step_size=None,
    time_half_bandwidth=4,
    n_tapers=7,
    frequency_range=None,
    standardize=True,
    calc_cpsd=False,
    calc_coh=True,
    return_weights=False,
    keepdims=False,
    n_jobs=1,
):
    """Calculates spectra for inferred states using a multitaper.

    Parameters
    ----------
    data : np.ndarray or list
        Time series data. Must have shape (n_sessions, n_samples,
        n_channels) or (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    alpha : np.ndarray or list, optional
        Inferred state probability time course. Must have shape
        (n_sessions, n_samples, n_states) or (n_samples, n_states).
    window_length : int, optional
        Length of the data segment to use to calculate spectra.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length` (no overlap).
    time_half_bandwidth : float, optional
        Parameter to control the resolution of the spectra.
    n_tapers : int, optional
        Number of tapers.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectra?
    calc_cpsd : bool, optional
        Should we return the cross spectra for :code:`psd`?
        If True, we force :code:`calc_coh` to False.
    calc_coh : bool, optional
        Should we also return the coherence spectra?
    return_weights : bool, optional
        Should we return the weights for session-specific spectra?
        This is useful for calculating a group average.
    keepdims : bool, optional
        Should we enforce a (n_sessions, n_states, ...) array is returned
        for :code:`psd` and :code:`coh`? If :code:`False`, we remove any
        dimension of length 1.
    n_jobs : int, optional
        Number of parallel jobs.

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences. Shape is (n_freq,).
    psd : np.ndarray
        Power spectra for each session and state. Shape is (n_sessions,
        n_states, n_channels, n_freq) if :code:`calc_cpsd=False`, otherwise
        the shape is (n_sessions, n_states, n_channels, n_channels, n_freq).
        Any axis of length 1 is removed if :code:`keepdims=False`.
    coh : np.ndarray
        Coherences for each session and state. Shape is (n_sessions, n_states,
        n_channels, n_channels, n_freq). Any axis of length 1 is removed if
        :code:`keepdims=False`. Only returned is :code:`calc_coh=True`.
    w : np.ndarray
        Weighting for session-specific spectra. Only returned if
        :code:`return_weights=True`. Shape is (n_sessions,).
    """

    # Validation
    if alpha is not None:
        if (isinstance(data, list) != isinstance(alpha, list)) or (
            isinstance(data, np.ndarray) != isinstance(alpha, np.ndarray)
        ):
            raise ValueError(
                f"data is type {type(data)} and alpha is type "
                f"{type(alpha)}. They must both be lists or numpy arrays."
            )

        if isinstance(data, list):
            # Check data and state time course for the same number of arrays
            # has been passed
            if len(data) != len(alpha):
                raise ValueError(
                    "A different number of arrays has been passed for "
                    f"data and alpha: len(data)={len(data)}, "
                    f"len(alpha)={len(alpha)}."
                )

            # Check the number of samples in data and alpha
            for i in range(len(alpha)):
                if alpha[i].shape[0] != data[i].shape[0]:
                    raise ValueError(
                        "items in data and alpha must have the same number of samples."
                    )

    elif isinstance(data, list):
        alpha = np.array([None] * len(data))

    if isinstance(data, np.ndarray):
        if alpha is not None and alpha.shape[0] != data.shape[0]:
            raise ValueError("data and alpha must have the number of samples.")

        if data.ndim == 2:
            data = [data]
            alpha = [alpha]

    if window_length is None:
        window_length = 2 * sampling_frequency
    window_length = int(window_length)

    if step_size is None:
        step_size = window_length
    step_size = int(step_size)

    if frequency_range is None:
        frequency_range = [0, sampling_frequency / 2]

    if calc_cpsd:
        calc_coh = False

    # Create keyword arguments to pass to the main function
    # that calculates spectra
    kwargs = []
    for d, a in zip(data, alpha):
        kwargs.append(
            {
                "data": d,
                "sampling_frequency": sampling_frequency,
                "alpha": a,
                "window_length": window_length,
                "step_size": step_size,
                "time_half_bandwidth": time_half_bandwidth,
                "n_tapers": n_tapers,
                "frequency_range": frequency_range,
                "standardize": standardize,
                "calc_cpsd": calc_cpsd,
                "calc_coh": calc_coh,
                "keepdims": True,
            }
        )

    # Main calculation of spectra
    if len(data) == 1:
        # We only have one array so we don't need to parallelise
        # the calculation
        _logger.info("Calculating spectra")
        results = [_multitaper(**kwargs[0])]

    elif n_jobs == 1:
        # We have multiple arrays but we're running in serial
        results = []
        for i in trange(len(data), desc="Calculating spectra"):
            results.append(_multitaper(**kwargs[i]))

    else:
        # Calculate spectra in parallel
        _logger.info("Calculating spectra")
        results = pqdm(
            kwargs,
            _multitaper,
            n_jobs=n_jobs,
            argument_type="kwargs",
        )

    # Unpack the results
    if calc_coh:
        psd = []
        coh = []
        for result in results:
            f, p, c = result
            psd.append(p)
            coh.append(c)
    else:
        psd = []
        for result in results:
            f, p = result
            psd.append(p)

    if not keepdims:
        # Remove any axes that are of length 1
        psd = np.squeeze(psd)
        if calc_coh:
            coh = np.squeeze(coh)

    # Weights for calculating a group-average spectrum
    n_samples = [d.shape[0] for d in data]
    w = np.array(n_samples) / np.sum(n_samples)

    if calc_coh:
        if return_weights:
            return f, psd, coh, w
        else:
            return f, psd, coh
    else:
        if return_weights:
            return f, psd, w
        else:
            return f, psd


def regression_spectra(
    data,
    alpha,
    sampling_frequency,
    window_length,
    step_size,
    n_sub_windows=1,
    frequency_range=None,
    standardize=True,
    calc_coh=True,
    return_weights=False,
    return_coef_int=False,
    keepdims=False,
    n_jobs=1,
):
    """Calculates mode-specific spectra by regressing a spectrogram with alpha.

    We use `_welch_spectrogram <https://osl-dynamics.readthedocs.io/en\
    /latest/autoapi/osl_dynamics/analysis/spectral/index.html#osl_dynamics\
    .analysis.spectral._welch_spectrogram>`_ to calculate the spectrogram.

    Parameters
    ----------
    data : np.ndarray or list
        Data to calculate a spectrogram for. Shape must be
        (n_sessions, n_samples, n_channels) or (n_samples, n_channels).
    alpha : np.ndarray or list
        Inferred mode mixing factors. Shape must be
        (n_sessions, n_samples, n_modes) or (n_samples, n_modes).
    sampling_frequency : float
        Sampling_frequency in Hz.
    window_length : int
        Number samples to use in the window to calculate a spectrum.
    step_size : int
        Step size for shifting the window.
    n_sub_windows : int, optional
        We split the window into a number of sub-windows and average the
        spectra for each sub-window. window_length must be divisible by
        :code:`n_sub_windows`.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectrogram?
    calc_coh : bool, optional
        Should we also return the coherence spectra?
    return_weights : bool, optional
        Should we return the weights for session-specific spectra?
        Useful for calculating the group average spectra.
    return_coef_int : bool, optional
        Should we return the regression coefficients and intercept
        separately for the mode PSDs?
    keepdims : bool, optional
        Should we enforce a (n_sessions, n_states, ...) array is returned for
        :code:`psd` and :code:`coh`? If :code:`False`, we remove any dimension
        of length 1.
    n_jobs : int, optional
        Number of parallel jobs.

    Returns
    -------
    f : np.ndarray
        Frequency axis. Shape is (n_freq).
    psd : np.ndarray
        Mode PSDs. A numpy array with shape (n_sessions, 2, n_modes, n_channels,
        n_freq) where :code:`axis=1` is the coefficients/intercept if
        :code:`return_coef_int=True`, otherwise shape is (n_sessions, n_modes,
        n_channels, n_freq). Any axis of length 1 is removed if
        :code:`keepdims=False`.
    coh : np.ndarray
        Mode coherences. Shape is (n_sessions, n_modes, n_channels, n_channels,
        n_freq). Any axis of length 1 is removed if :code:`keepdims=False`.
    w : np.ndarray
        Weight for each session-specific PSD. Shape is (n_sessions,).
        Only returned if :code:`return_weights=True`.
    """

    # Validation
    if (isinstance(data, list) != isinstance(alpha, list)) or (
        isinstance(data, np.ndarray) != isinstance(alpha, np.ndarray)
    ):
        raise ValueError(
            f"data is type {type(data)} and alpha is type "
            f"{type(alpha)}. They must both be lists or numpy arrays."
        )

    if isinstance(data, list):
        # Check data and state time course for the same number of arrays
        # has been passed
        if len(data) != len(alpha):
            raise ValueError(
                "A different number of arrays has been passed for "
                f"data and alpha: len(data)={len(data)}, "
                f"len(alpha)={len(alpha)}."
            )

        # Check the number of samples in data and alpha
        for i in range(len(alpha)):
            if alpha[i].shape[0] != data[i].shape[0]:
                raise ValueError(
                    "items in data and alpha must have the number of samples."
                )

    if isinstance(data, np.ndarray):
        if alpha.shape[0] != data.shape[0]:
            raise ValueError("data and alpha must have the number of samples.")

        if data.ndim == 2:
            data = [data]
            alpha = [alpha]

    # Ensure data and alpha are float32
    data = [d.astype(np.float32) for d in data]
    alpha = [a.astype(np.float32) for a in alpha]

    if frequency_range is None:
        frequency_range = [0, sampling_frequency / 2]

    if window_length % n_sub_windows != 0:
        raise ValueError("window_length must be divisible by n_sub_windows.")

    # Standardise before calculating the spectrogram
    if standardize:
        data = [(d - np.mean(d, axis=0)) / np.std(d, axis=0) for d in data]

    # Create keyword arguments to pass to the main function
    # that calculates spectra
    kwargs = []
    for d, a in zip(data, alpha):
        kwargs.append(
            {
                "data": d,
                "alpha": a,
                "sampling_frequency": sampling_frequency,
                "window_length": window_length,
                "step_size": step_size,
                "frequency_range": frequency_range,
                "calc_cpsd": calc_coh,
                "n_sub_windows": n_sub_windows,
            }
        )

    # Main calculation of spectra
    if len(data) == 1:
        # We only have one array so we don't need to parallelise the
        # calculation
        _logger.info("Calculating spectra")
        results = [_regress_welch_spectrogram(**kwargs[0])]

    elif n_jobs == 1:
        # We have multiple arrays but we're running in serial
        results = []
        for i, kws in enumerate(kwargs):
            _logger.info(f"Calculating spectra {i}")
            results.append(_regress_welch_spectrogram(**kws))
    else:
        # Calculate a time-varying PSD and regress to get the mode PSDs
        _logger.info("Calculating spectra")
        results = pqdm(
            kwargs,
            _regress_welch_spectrogram,
            n_jobs=n_jobs,
            argument_type="kwargs",
        )

    # Unpack results
    Pj = []
    for result in results:
        f, coefs, intercept = result
        Pj.append([coefs, [intercept] * coefs.shape[0]])
    Pj = np.array(Pj)

    # Weights for calculating a group-average spectrum
    n_samples = [d.shape[0] for d in data]
    w = np.array(n_samples) / np.sum(n_samples)

    if not calc_coh:
        if not return_coef_int:
            # Sum coefficients and intercept
            Pj = np.sum(Pj, axis=1)

        if return_weights:
            return f, np.squeeze(Pj), w
        else:
            return f, np.squeeze(Pj)

    # Number of channels and freqency bins
    n_channels = data[0].shape[1]
    n_modes = alpha[0].shape[1]
    n_freq = Pj.shape[-1]

    # Indices of the upper triangle of an n_channels by n_channels array
    m, n = np.triu_indices(n_channels)

    # Number of arrays
    n_sessions = len(data)

    # Create a n_channels by n_channels array
    P = np.empty(
        [n_sessions, 2, n_modes, n_channels, n_channels, n_freq],
        dtype=np.complex64,
    )
    for i in range(n_sessions):
        for j in range(2):  # j=0 is the coefficients and j=1 is the intercepts
            P[i, j][:, m, n] = Pj[i, j]
            P[i, j][:, n, m] = np.conj(Pj[i, j])

    # PSDs and coherences for each mode
    psd = np.empty(
        [n_sessions, 2, n_modes, n_channels, n_freq],
        dtype=np.float32,
    )
    coh = np.empty(
        [n_sessions, n_modes, n_channels, n_channels, n_freq],
        dtype=np.float32,
    )
    for i in range(n_sessions):
        # PSDs
        p = P[i, :, :, range(n_channels), range(n_channels)].real
        psd[i] = np.rollaxis(p, axis=0, start=3)

        # Calculate coherence
        for k in range(n_channels):
            for l in range(n_channels):
                Pxy = P[i, 0, :, k, l]  # regression coefficients
                Pxx = P[i, 1, :, k, k].real  # regression intercept
                Pyy = P[i, 1, :, l, l].real  # regression intercept
                coh[i, :, k, l] = np.abs(Pxy) / np.sqrt(Pxx * Pyy)

    if not return_coef_int:
        # Sum coefficients and intercept
        psd = np.sum(psd, axis=1)

    if not keepdims:
        # Remove any axes that are of length 1
        psd = np.squeeze(psd)
        coh = np.squeeze(coh)

    if return_weights:
        return f, psd, coh, w
    else:
        return f, psd, coh


def rescale_regression_coefs(
    psd,
    alpha,
    window_length,
    step_size,
    n_sub_windows=1,
):
    """Rescales regression coefficients with maximum regressor values.

    Parameters
    ----------
    psd : np.ndarray
        Mode PSDs. Shape must be (n_sessions, 2, n_modes, n_channels, n_freq).
    alpha : np.ndarray or list
        Inferred mode mixing coefficients. Shape must be (n_sessions, n_samples,
        n_modes) or (n_samples, n_modes).
    window_length : int
        Number samples used as the window to calcualte a PSD.
    step_size : int
        Step size used for shifting the window.
    n_sub_windows : int, optional
        Number of sub-windows used to average the spectra. :code:`window_length`
        must be divisible by :code:`n_sub_windows`.

    Returns
    -------
    psd : np.ndarray
        Mode PSDs rescaled with maximum regressor values.
        Shape is identical to the input.
    """

    # Validation
    if psd.ndim == 4:
        psd = psd[np.newaxis, ...]

    if psd.ndim != 5:
        raise ValueError(
            "psd must have shape (n_sessions, 2, n_modes, n_channels, n_freqs)."
        )

    if isinstance(alpha, np.ndarray):
        alpha = [alpha]

    if len(psd) != len(alpha):
        raise ValueError(
            "A different number of arrays has been passed for "
            f"psd and alpha: len(psd)={len(psd)}, "
            f"len(alpha)={len(alpha)}."
        )

    for i in range(len(alpha)):
        if alpha[i].shape[1] != psd[i].shape[1]:
            raise ValueError("psd and alpha must have same number of modes.")

    # Prevent in-place operation
    psd = psd.copy()

    # Number of arrays
    n_sessions = len(alpha)

    # Window alphas to match the windowing of STFT
    for i, a in enumerate(alpha):
        alpha[i] = _window_mean(
            a,
            "hann",
            window_length,
            step_size=step_size,
            n_sub_windows=n_sub_windows,
        )

    # Normalize the alphas
    alpha = [(a - np.mean(a, axis=0)) / np.std(a, axis=0) for a in alpha]

    # Get maximum regressor values
    max_alpha = [np.max(a, axis=0) for a in alpha]

    # Rescale the regression coefficients
    for n in range(n_sessions):
        psd[n, 0] *= np.expand_dims(max_alpha[n], axis=(1, 2))

    return psd


def coherence_spectra(cpsd, keepdims=False):
    r"""Calculate coherence from cross power spectral densities.

    We calculate the coherence as:

    .. math::
        C_{xy}(f) = \frac{|P_{xy}(f)|}{\sqrt{P_{xx}(f) P_{yy}(f)}}

    where :math:`x,y` are different channels.

    Parameters
    ----------
    cpsd : np.ndarray
        Cross power spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    keepdims: bool, optional
        Should we squeeze any axis of length 1?

    Returns
    -------
    coh : np.ndarray
        Coherence spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    """
    n_channels = cpsd.shape[-2]

    # Calculate coherency spectrum
    coh = np.empty_like(cpsd, dtype=np.float32)
    for j in range(n_channels):
        for k in range(n_channels):
            Pxy = cpsd[..., j, k, :]
            Pxx = cpsd[..., j, j, :].real
            Pyy = cpsd[..., k, k, :].real
            coh[..., j, k, :] = abs(Pxy) / np.sqrt(Pxx * Pyy)

    # Zero nan values
    coh = np.nan_to_num(coh)

    if not keepdims:
        coh = np.squeeze(coh)

    return coh


def partial_coherence_spectra(cpsd, keepdims=False):
    r"""Calculate partial coherence from cross power spectral densities.

    We calculate the partial coherence as:

    .. math::
        PC_{xy}(f) = \
            \frac{|P^{-1}_{xy}(f)|}{\sqrt{P^{-1}_{xx}(f) P^{-1}_{yy}(f)}}

    where :math:`x,y` are different channels and :math:`P^{-1}(f)`
    is the inverse of the cross spectral density matrix.

    Parameters
    ----------
    cpsd : np.ndarray
        Cross power spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    keepdims: bool, optional
        Should we squeeze any axis of length 1?

    Returns
    -------
    pcoh : np.ndarray
        Partial coherence spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    """
    # Calculate inverse cross spectra
    # (Need to move axis for compatibility with np.linalg.pinv)
    cpsd = np.moveaxis(cpsd, -1, 0)
    icpsd = np.linalg.inv(cpsd)
    icpsd = np.moveaxis(icpsd, 0, -1)

    # The partial coherence calculation is the same as the
    # coherence calculation except we input the inverse cross
    # spectral densities
    return coherence_spectra(icpsd, keepdims=keepdims)


def partial_directed_coherence_spectra(
    f,
    cpsd,
    sampling_frequency,
    n_iterations=50,
    tol=1e-7,
    keepdims=False,
):
    r"""Calculate partial directed coherence from cross power spectra.

    We calculate the partial directed coherence as:

    .. math::
        PDC_{ij}(f) = \frac{|A_{ij}(f)|}{\sqrt{\sum_m |A_{mj}(f)|^2}}

    where :math:`i,j` are different channels and :math:`A(f) = H^{-1}(f)`
    is the inverse of the transfer function, which is calculated by
    factorizing the cross spectral density matrix.

    By default this function will use all cores available.

    Parameters
    ----------
    f : np.ndarray
        Frequency axis. Shape is (n_freq,).
    cpsd : np.ndarray
        Cross power spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    sampling_frequency : float
        Sampling frequency (Hz).
    n_iterations : int, optional
        Number of iterations in the Wilson factorization of the cross
        spectra.
    tol : float, optional
        Tolerance for the Wilson factorization of the cross spectra.
    keepdims: bool, optional
        Should we squeeze any axis of length 1?

    Returns
    -------
    pdc : np.ndarray
        Partial directed coherence spectra.
        Shape is (..., n_channels, n_channels, n_freq).
    """

    def _plus_operator(g):
        # Helper function for _wilson_factorization
        N = g.shape[0] // 2 + 1
        gam = np.fft.ifft(g, axis=0)
        beta0 = 0.5 * gam[0]
        gamp = gam.copy()
        gamp[0] = np.triu(beta0)
        gamp[N:] = 0
        return np.fft.fft(gamp, axis=0)

    def _wilson_factorization(
        S, freq=f, fs=sampling_frequency, Niterations=n_iterations, tol=tol
    ):
        # Algorithm for the Wilson Factorization of the spectral matrix. From:
        # https://github.com/ViniciusLima94/pyGC/blob/master/pygc/non_parametric.py
        #
        # Args:
        # - Cross spectra: S (frequencies, channels, channels)
        #
        # Returns:
        # - Transfer function: H (frequencies, channels, channels)
        m = S.shape[-1]
        N = freq.shape[0] - 1
        Sarr = np.zeros([2 * N, m, m], dtype=complex)
        Sarr[:N] = S[:N]
        Sarr[N:] = S[1:].transpose(0, 2, 1)[::-1]
        gam = np.fft.ifft(Sarr, axis=0).real
        h = np.linalg.cholesky(gam[0]).T
        psi = np.tile(h[np.newaxis, ...], (2 * N, 1, 1)).astype(complex)
        I = np.eye(m)[np.newaxis, ...]
        for _ in range(Niterations):
            inv_psi = np.linalg.inv(psi)
            g = inv_psi @ Sarr @ np.conj(inv_psi).transpose(0, 2, 1) + I
            gp = _plus_operator(g)
            psiold = psi.copy()
            psi = psi @ gp
            psierr = np.mean(np.linalg.norm(psi - psiold, 1, axis=1)) / (2 * N)
            if psierr < tol:
                break
        A = np.fft.ifft(psi, axis=0).real
        inv_A0 = np.linalg.inv(A[0])
        return psi[: N + 1] @ inv_A0

    def _calc_pdc(cpsd):
        # Calculate partial directed coherence from a single cross spectrum.
        #
        # Args:
        # - cpsd (n_channels, n_channels, n_freq) array
        #
        # Returns:
        # - pdc (n_channels, n_channels, n_freq) array
        S = np.moveaxis(cpsd, -1, 0)
        H = _wilson_factorization(S)
        A = np.linalg.inv(H)
        pdc = np.abs(A) / np.sqrt(np.sum(np.abs(A) ** 2, axis=1, keepdims=True))
        return np.moveaxis(pdc, 0, -1).astype(np.float32)

    # _calc_pdc works on 3D data, we flatten the input here
    original_shape = cpsd.shape
    flattened_cpsd = cpsd.reshape((-1, *original_shape[-3:]))

    # Create keyword arguments to pass to _calc_pdc
    kwargs = [{"cpsd": S} for S in flattened_cpsd]

    # Main calculation
    if len(kwargs) == 1:
        _logger.info("Calculating PDC")
        results = [_calc_pdc(**kwargs[0])]
    else:
        results = []
        for i in trange(len(kwargs), desc="Calculating PDC"):
            results.append(_calc_pdc(**kwargs[i]))

    # Put the results back into the original shape
    pdc = np.array(results).reshape(original_shape)

    if not keepdims:
        # Remove any axes that are of length 1
        pdc = np.squeeze(pdc)

    return pdc


def decompose_spectra(
    coherences,
    n_components,
    max_iter=50000,
    random_state=None,
    verbose=0,
):
    """Uses non-negative matrix factorization to decompose spectra.

    Parameters
    ----------
    coherences : np.ndarray
        Coherences spectra.
        Shape must be (..., n_channels, n_channels, n_freq,).
    n_components : int
        Number of spectral components to fit.
    max_iter : int, optional
        Maximum number of iterations in
        `sklearn.decomposion.non_negative_factorization
        <https://scikit-learn.org/stable/modules/generated/sklearn\
        .decomposition.non_negative_factorization.html>`_.
    random_state : int, optional
        Seed for the random number generator.
    verbose : int, optional
        Show verbose? (:code:`1`) yes, (:code:`0`) no.

    Returns
    -------
    components : np.ndarray
        Spectral components. Shape is (n_components, n_freq).
    """
    _logger.info("Performing spectral decomposition")

    # Validation
    error_message = (
        "coherences must be a numpy array with shape "
        "(n_channels, n_channels, n_freq), "
        "(n_modes, n_channels, n_channels, n_freq) or "
        "(n_sessions, n_modes, n_channels, n_channels, n_freq)."
    )
    coherences = array_ops.validate(
        coherences,
        correct_dimensionality=5,
        allow_dimensions=[3, 4],
        error_message=error_message,
    )

    # Number of arrays, modes, channels and frequency bins
    n_sessions, n_modes, n_channels, n_channels, n_freq = coherences.shape

    # Indices of the upper triangle of the
    # [n_channels, n_channels, n_freq] sub-array
    i, j = np.triu_indices(n_channels, 1)

    # Concatenate coherences for each array and mode
    # and only keep the upper triangle
    coherences = coherences[:, :, i, j].reshape(-1, n_freq)

    # Perform non-negative matrix factorisation
    _, components, _ = non_negative_factorization(
        coherences,
        n_components=n_components,
        init=None,
        max_iter=max_iter,
        random_state=random_state,
        verbose=verbose,
    )

    # Order the weights and components in ascending frequency
    order = np.argsort(components.argmax(axis=1))
    components = components[order]

    return components


def mar_spectra(coeffs, covs, sampling_frequency, n_freq=512):
    """Calculates cross power spectral densities from MAR model parameters.

    Parameters
    ----------
    coeffs : np.ndarray
        MAR coefficients. Shape must be (n_modes, n_channels, n_channels) or
        (n_channels, n_channels) or (n_channels,).
    covs : np.ndarray
        MAR covariances. Shape must be (n_modes, n_channels, n_channels) or
        (n_channels, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    n_freq : int, optional
        Number of frequency bins in the cross power spectral density to
        calculate.

    Returns
    -------
    f : np.ndarray
        Frequency axis. Shape is (n_freq,).
    P : np.ndarray
        Cross power spectral densities. Shape is (n_freq, n_modes, n_channels,
        n_channels) or (n_freq, n_channels, n_channels).
    """
    # Validation
    if covs.shape[-1] != covs.shape[-2]:
        if covs.ndim == 2:
            covs = [np.diag(c) for c in covs]
        else:
            covs = np.diag(covs)
    error_message = (
        "covs must be a numpy array with shape "
        "(n_modes, n_channels, n_channels), "
        "(n_channels, n_channels) or (n_channels,)."
    )
    covs = array_ops.validate(
        covs,
        correct_dimensionality=3,
        allow_dimensions=[2],
        error_message=error_message,
    )
    error_message = (
        "coeffs must be a numpy array with shape "
        "(n_modes, n_lags, n_channels, n_channels), "
        "(n_lags, n_channels, n_channels)."
    )
    coeffs = array_ops.validate(
        coeffs,
        correct_dimensionality=4,
        allow_dimensions=[3],
        error_message=error_message,
    )

    n_modes = coeffs.shape[0]
    n_lags = coeffs.shape[1]
    n_channels = coeffs.shape[-1]

    # Frequencies to evaluate the PSD at
    f = np.arange(0, sampling_frequency / 2, sampling_frequency / (2 * n_freq))

    # z-transform of the coefficients
    A = np.zeros([n_freq, n_modes, n_channels, n_channels], dtype=np.complex64)
    for i in range(n_freq):
        for l in range(n_lags):
            z = np.exp(-1j * (l + 1) * 2 * np.pi * f[i] / sampling_frequency)
            A[i] += coeffs[:, l] * z

    # Transfer function
    I = np.identity(n_channels)[np.newaxis, np.newaxis, ...]
    e = 1e-6 * I
    H = np.linalg.inv(I - A + e)

    # Cross PSDs
    P = H @ covs[np.newaxis, ...] @ np.transpose(np.conj(H), axes=[0, 1, 3, 2])

    return f, np.squeeze(P)


def autocorr_to_spectra(
    autocorr_func,
    sampling_frequency,
    nfft=64,
    frequency_range=None,
):
    """Calculates spectra from the autocorrelation function.

    The power spectrum of each mode is calculated as the Fourier transform of
    the auto-correlation function. Coherences are calculated from the power
    spectra.

    Parameters
    ----------
    autocorr_func : np.ndarray
        Autocorrelation functions. Shape must be (n_channels, n_channels,
        n_lags) or (n_modes, n_channels, n_channels, n_lags).
    sampling_frequency : float
        Sampling frequency in Hz.
    nfft : int, optional
        Number of data points in the FFT. The auto-correlation function will
        only have :code:`2 * (n_embeddings + 2) - 1` data points. We pad the
        auto-correlation function with zeros to have :code:`nfft` data points
        if the number of data points in the auto-correlation function is less
        than :code:`nfft`.
    frequency_range : list, optional
        Minimum and maximum frequency to keep (Hz).

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences. Shape is (n_freq,).
    psd : np.ndarray
        Power (or cross) spectra calculated for each mode. Shape is
        (n_channels, n_channels, n_freq) or (n_modes, n_channels, n_channels,
        n_freq).
    coh : np.ndarray
        Coherences calculated for each mode. Shape is (n_channels, n_channels,
        n_freq) or (n_modes, n_channels, n_channels, n_freq).
    """
    # Validation
    error_message = (
        "autocorrelation_functions must be of shape (n_channels, n_channels, "
        "n_lags) or (n_modes, n_channels, n_channels, n_lags)."
    )
    autocorr_func = array_ops.validate(
        autocorr_func,
        correct_dimensionality=4,
        allow_dimensions=[3],
        error_message=error_message,
    )
    if frequency_range is None:
        frequency_range = [0, sampling_frequency / 2]

    _logger.info("Calculating power spectra")

    # Number of data points in the autocorrelation function and FFT
    n_lags = autocorr_func.shape[-1]
    nfft = max(nfft, 2 ** nextpow2(n_lags))

    # Calculate the argments to keep for the given frequency range
    f = np.arange(
        0,
        sampling_frequency / 2,
        sampling_frequency / nfft,
    )
    [min_arg, max_arg] = get_frequency_args_range(f, frequency_range)
    f = f[min_arg:max_arg]

    # Calculate cross power spectra as the Fourier transform of the
    # auto/cross-correlation function
    psd = np.fft.fft(autocorr_func, nfft)
    psd = psd[..., min_arg:max_arg]
    psd = abs(psd)

    # Normalise the power spectra
    psd /= nfft**2

    # Coherences for each mode
    coh = coherence_spectra(psd, keepdims=True)

    return f, np.squeeze(psd), np.squeeze(coh)


def get_frequency_args_range(frequencies, frequency_range):
    """Get min/max indices of a range in a frequency axis.

    Parameters
    ----------
    frequencies : np.ndarray
        1D numpy array containing a Frequency axis.
    frequency_range : list of len 2
        Minimum and maximum frequency.

    Returns
    -------
    args_range : list of len 2
        Indices for minimum and maximum frequency.
    """
    f_min_arg = np.argwhere(frequencies >= frequency_range[0])[0, 0]
    f_max_arg = np.argwhere(frequencies <= frequency_range[1])[-1, 0]
    if f_max_arg <= f_min_arg:
        raise ValueError("Cannot select requested frequency range.")
    return [f_min_arg, f_max_arg]


def _welch_spectrogram(
    data,
    sampling_frequency,
    window_length=None,
    step_size=None,
    frequency_range=None,
    calc_cpsd=True,
    n_sub_windows=1,
):
    """Calculates a spectogram (time-varying power spectral density)
    using Welch's method.

    Steps:

    1. Segment data into overlapping windows.
    2. Multiply each window by a Hann tapering function.
    3. Calculate a periodogram for each window.

    We use the same scaling for the power spectra as SciPy (i.e. when
    you use :code:`scipy.signal.periodogram(..., scaling="density")`).

    Parameters
    ----------
    data : np.ndarray
        Data to calculate the spectrogram for.
        Shape must be (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    window_length : int, optional
        Number of data points to use when calculating the periodogram.
        Defaults to :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length // 2` (50% overlap).
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    calc_cpsd : bool, optional
        Should we calculate cross spectra?
    n_sub_windows : int, optional
        We split the window into a number of sub-windows and average the
        spectra for each sub-window. window_length must be divisible by
        :code:`n_sub_windows`.

    Returns
    -------
    t : np.ndarray
        Time axis.
    f : np.ndarray
        Frequency axis.
    P : np.ndarray
        Spectrogram. 3D numpy array with shape (time, channels *
        (channels + 1) / 2, freq) if :code:`calc_cpsd=True`, otherwise
        it is (time, channels, freq).
    """
    n_samples, n_channels = data.shape

    # Validation
    if window_length is None:
        window_length = 2 * sampling_frequency
    window_length = int(window_length)

    if step_size is None:
        step_size = window_length // 2
    step_size = int(step_size)

    # First pad the data so we have enough data points to estimate the
    # periodogram for time points at the start/end of the data
    data = np.pad(data, window_length // 2)[
        :, window_length // 2 : window_length // 2 + n_channels
    ]

    # Window to apply to the data before calculating the Fourier transform
    window = signal.get_window("hann", (window_length // n_sub_windows))

    # Number of data points in the FFT
    nfft = max(256, window_length // n_sub_windows)

    # Time and frequency axis
    t = np.arange(n_samples) / sampling_frequency
    f = np.arange(nfft // 2) * sampling_frequency / nfft

    # Only keep a particular frequency range
    [min_arg, max_arg] = get_frequency_args_range(f, frequency_range)
    f = f[min_arg:max_arg]

    # Number of frequency bins
    n_freq = max_arg - min_arg

    # Indices of an upper triangle of an n_channels by n_channels array
    m, n = np.triu_indices(n_channels)

    # Indices of time points to calculate a periodogram for
    time_indices = range(0, n_samples, step_size)
    n_windows = n_samples // step_size

    if calc_cpsd:
        # Calculate cross periodograms for each segment of the data
        P = np.empty(
            [n_windows, n_channels * (n_channels + 1) // 2, n_freq],
            dtype=np.complex64,
        )
        XY_sub_window = np.empty(
            [n_sub_windows, n_channels * (n_channels + 1) // 2, n_freq],
            dtype=np.complex64,
        )
        for i in range(n_windows):
            # Data in the window
            j = time_indices[i]
            x_window = data[j : j + window_length].T

            for k in range(n_sub_windows):
                # Data in the sub-window with the windowing function applied
                x_sub_window = (
                    x_window[
                        :,
                        k
                        * window_length
                        // n_sub_windows : (k + 1)
                        * window_length
                        // n_sub_windows,
                    ]
                    * window[np.newaxis, ...]
                )

                # Calculate cross spectra for the sub-window
                X = np.fft.fft(x_sub_window, nfft)
                X = X[..., min_arg:max_arg]
                XY = np.conj(X)[np.newaxis, :, :] * X[:, np.newaxis, :]
                XY_sub_window[k] = XY[m, n]

            # Average the cross spectra for each sub-window
            P[i] = np.mean(XY_sub_window, axis=0)

    else:
        # Calculate the periodogram for each segment of the data
        P = np.empty([n_windows, n_channels, n_freq], dtype=np.float32)
        XX_sub_window = np.empty(
            [n_sub_windows, n_channels, n_freq],
            dtype=np.float32,
        )
        for i in range(n_windows):
            # Data in the window
            j = time_indices[i]
            x_window = data[j : j + window_length].T

            for k in range(n_sub_windows):
                # Data in the sub-window with the windowing function applied
                x_sub_window = (
                    x_window[
                        :,
                        k
                        * window_length
                        // n_sub_windows : (k + 1)
                        * window_length
                        // n_sub_windows,
                    ]
                    * window[np.newaxis, ...]
                )

                # Calculate PSD for the sub-window
                X = np.fft.fft(x_sub_window, nfft)
                X = X[..., min_arg:max_arg]
                XX_sub_window[k] = np.real(np.conj(X) * X)

            # Average the cross spectra for each sub-window
            P[i] = np.mean(XX_sub_window, axis=0)

    # Scaling for the periodograms (we use the same scaling as SciPy)
    P *= 2 / (sampling_frequency * np.sum(window**2))

    return t, f, P


def _welch(
    data,
    sampling_frequency,
    alpha=None,
    window_length=None,
    step_size=None,
    frequency_range=None,
    standardize=True,
    calc_coh=False,
    keepdims=False,
):
    """Calculate a (cross) power spectrum using Welch's method.

    This function first calculates a spectrogram using
    `_welch_spectrogram <https://osl-dynamics.readthedocs.io/en/latest\
    /autoapi/osl_dynamics/analysis/spectral/index.html#osl_dynamics.analysis\
    .spectral._welch_spectrogram>`_ then takes the time average.

    The scaling for the power spectra calculated by this function
    matches SciPy (:code:`scipy.signal.welch`).

    Parameters
    ----------
    data : np.ndarray
        Time series data. Must have shape (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    alpha : np.ndarray, optional
        Inferred state probability time course.
        Must have shape (n_samples, n_states).
    window_length : int, optional
        Length of the data segment to use to calculate spectra.
        If None, we use :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length // 2` (50% overlap).
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectra?
    calc_coh : bool, optional
        Should we calculate the coherence between channels?
    keepdims : bool, optional
        Should we squeeze any axes of length 1?

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences.
        Shape is (n_freq,).
    psd : np.ndarray
        Power spectra. Shape is (..., n_channels, n_freq).
    coh : np.ndarray
        Coherences spectra.
        Shape is (..., n_channels, n_channels, n_freq).
        Only returned is :code:`calc_coh=True`.
    """
    n_samples, n_channels = data.shape

    if alpha is None:
        alpha = np.ones([n_samples, 1], dtype=np.float32)
    n_states = alpha.shape[-1]

    # Indices for the upper triangle of a channels by channels matrix
    m, n = np.triu_indices(n_channels)

    # Standardise before calculating spectra
    if standardize:
        data = (data - np.mean(data, axis=0)) / np.std(data, axis=0)

    # Rescaling for the PSD to account for how much time each
    # state was active. Note, we using the same scaling as the
    # the multitaper calculation in the HMM-MAR toolbox
    sum_alpha2 = np.sum(alpha**2, axis=0)
    scaling = n_samples / sum_alpha2

    psd = []
    coh = []
    for i in range(n_states):
        # Calculate spectrogram for this state's data
        x = data * alpha[..., i][..., np.newaxis]
        _, f, p = _welch_spectrogram(
            data=x,
            sampling_frequency=sampling_frequency,
            window_length=window_length,
            step_size=step_size,
            frequency_range=frequency_range,
            calc_cpsd=calc_coh,
        )

        # Average over the time dimension and rescale
        p = np.mean(p, axis=0)
        p *= scaling[i]

        if calc_coh:
            # Create a channels by channels matrix for cross PSDs
            n_freq = p.shape[-1]
            cpsd = np.empty(
                [n_channels, n_channels, n_freq],
                dtype=np.complex64,
            )
            cpsd[m, n] = p
            cpsd[n, m] = np.conj(p)

            # Unpack PSDs
            psd.append(cpsd[range(n_channels), range(n_channels)].real)

            # Calculate coherence
            coh.append(coherence_spectra(cpsd))
        else:
            psd.append(p)

    # Check for nans
    if np.isnan(psd).any():
        psd = np.nan_to_num(psd)  # zero out nan values
        _logger.warn(
            "PSD contains NaN values. This may indicate a potentially "
            "poor HMM fit. You should consider running the model again "
            "or selecting the model run with the lowest free energy after "
            "training multiple times."
        )

    if not keepdims:
        # Squeeze any axes of length 1
        psd = np.squeeze(psd)
        if calc_coh:
            coh = np.squeeze(coh)

    if calc_coh:
        return f, psd, coh
    else:
        return f, psd


def _multitaper_spectrogram(
    data,
    sampling_frequency,
    window_length=None,
    step_size=None,
    frequency_range=None,
    calc_cpsd=True,
    n_sub_windows=1,
    time_half_bandwidth=4,
    n_tapers=7,
):
    """Calculates a spectogram (time-varying power spectral density)
    using a multitaper.

    Steps:

    1. Segment data into overlapping windows.
    2. Multiply each window by a number of tapering functions.
    3. Calculate a periodogram for each tapered window.
    4. Average over tapers.

    Parameters
    ----------
    data : np.ndarray
        Data to calculate the spectrogram for.
        Shape must be (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    window_length : int, optional
        Number of data points to use when calculating the periodogram.
        Defaults to :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length` (no overlap).
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    calc_cpsd : bool, optional
        Should we calculate cross spectra?
    n_sub_windows : int, optional
        We split the window into a number of sub-windows and average the
        spectra for each sub-window. window_length must be divisible by
        :code:`n_sub_windows`.
    time_half_bandwidth : float, optional
        Parameter to control the resolution of the spectra.
    n_tapers : int, optional
        Number of tapers.

    Returns
    -------
    t : np.ndarray
        Time axis.
    f : np.ndarray
        Frequency axis.
    P : np.ndarray
        Spectrogram. 3D numpy array with shape (time, channels *
        (channels + 1) / 2, freq) if :code:`calc_cpsd=True`, otherwise
        it is (time, channels, freq).
    """
    n_samples, n_channels = data.shape

    # Validation
    if window_length is None:
        window_length = 2 * sampling_frequency
    window_length = int(window_length)

    if step_size is None:
        step_size = window_length
    step_size = int(step_size)

    # First pad the data so we have enough data points to estimate the
    # periodogram for time points at the start/end of the data
    data = np.pad(data, window_length // 2)[
        :, window_length // 2 : window_length // 2 + n_channels
    ]

    # Tapering windows to apply to the data before calculating the
    # Fourier transform
    tapers = signal.windows.dpss(
        window_length,
        NW=time_half_bandwidth,
        Kmax=n_tapers,
    )

    # Number of data points in the FFT
    nfft = window_length // n_sub_windows

    # Time and frequency axis
    t = np.arange(n_samples) / sampling_frequency
    f = np.arange(nfft // 2) * sampling_frequency / nfft

    # Only keep a particular frequency range
    [min_arg, max_arg] = get_frequency_args_range(f, frequency_range)
    f = f[min_arg:max_arg]

    # Number of frequency bins
    n_freq = max_arg - min_arg

    # Indices of an upper triangle of an n_channels by n_channels array
    m, n = np.triu_indices(n_channels)

    # Indices of time points to calculate a periodogram for
    time_indices = range(0, n_samples, step_size)
    n_windows = n_samples // step_size

    if calc_cpsd:
        # Calculate cross PSDs for each segment of the data
        P = np.empty(
            [n_windows, n_channels * (n_channels + 1) // 2, n_freq],
            dtype=np.complex64,
        )
        XY_sub_window = np.empty(
            [n_sub_windows, n_channels * (n_channels + 1) // 2, n_freq],
            dtype=np.complex64,
        )
        for i in range(n_windows):
            # Data in the window
            j = time_indices[i]
            x_window = data[j : j + window_length].T

            for k in range(n_sub_windows):
                # Data in the sub-window
                x_sub_window = x_window[
                    :,
                    k
                    * window_length
                    // n_sub_windows : (k + 1)
                    * window_length
                    // n_sub_windows,
                ]

                # Apply tapers
                x_sub_window = x_sub_window[np.newaxis, :, :] * tapers[:, np.newaxis, :]

                # Fourier transform
                X = np.fft.fft(x_sub_window, nfft)
                X = X[..., min_arg:max_arg]

                # Calculate cross spectra for the sub-window
                XY = np.conj(X)[:, :, np.newaxis, :] * X[:, np.newaxis, :, :]
                XY_sub_window[k] = np.mean(XY[:, m, n], axis=0)

            # Average the cross spectra for each sub-window
            P[i] = np.mean(XY_sub_window, axis=0)

    else:
        # Calculate PSDs for each segment of the data
        P = np.empty([n_windows, n_channels, n_freq], dtype=np.float32)
        XX_sub_window = np.empty(
            [n_sub_windows, n_channels, n_freq],
            dtype=np.float32,
        )
        for i in range(n_windows):
            # Data in the window
            j = time_indices[i]
            x_window = data[j : j + window_length].T

            for k in range(n_sub_windows):
                # Data in the sub-window
                x_sub_window = x_window[
                    :,
                    k
                    * window_length
                    // n_sub_windows : (k + 1)
                    * window_length
                    // n_sub_windows,
                ]

                # Apply tapers
                x_sub_window = x_sub_window[np.newaxis, :, :] * tapers[:, np.newaxis, :]

                # Fourier transform
                X = np.fft.fft(x_sub_window, nfft)
                X = X[..., min_arg:max_arg]

                # Calculate spectra for the sub-window
                XX = np.real(np.conj(X) * X)
                XX_sub_window[k] = np.mean(XX, axis=0)

            # Average the spectra for each sub-window
            P[i] = np.mean(XX_sub_window, axis=0)

    # Scaling for the multitapers
    P *= 2 / sampling_frequency

    return t, f, P


def _multitaper(
    data,
    sampling_frequency,
    alpha=None,
    window_length=None,
    step_size=None,
    time_half_bandwidth=4,
    n_tapers=7,
    frequency_range=None,
    standardize=True,
    calc_cpsd=False,
    calc_coh=True,
    keepdims=False,
):
    """Calculate a multitaper.

    This function first calculates a spectrogram using
    `_multitaper_spectrogram <https://osl-dynamics.readthedocs.io/en/latest\
    /autoapi/osl_dynamics/analysis/spectral/index.html#osl_dynamics.analysis\
    .spectral._multitaper_spectrogram>`_ then takes the time average.

    Parameters
    ----------
    data : np.ndarray
        Time series data. Must have shape (n_samples, n_channels).
    sampling_frequency : float
        Sampling frequency in Hz.
    alpha : np.ndarray, optional
        Inferred state probability time course.
        Must have shape (n_samples, n_states).
    window_length : int, optional
        Length of the data segment to use to calculate spectra.
        If None, we use :code:`2 * sampling_frequency`.
    step_size : int, optional
        Step size for shifting the window.
        Defaults to :code:`window_length` (no overlap).
    time_half_bandwidth : float, optional
        Parameter to control the resolution of the spectra.
    n_tapers : int, optional
        Number of tapers.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    standardize : bool, optional
        Should we standardize the data before calculating the spectra?
    calc_cpsd : bool, optional
        Should we return the cross spectra for :code:`psd`?
    calc_coh : bool, optional
        Should we calculate the coherence between channels?
    keepdims : bool, optional
        Should we squeeze any axes of length 1?

    Returns
    -------
    f : np.ndarray
        Frequencies of the power spectra and coherences.
        Shape is (n_freq,).
    psd : np.ndarray
        Power spectra. Shape is (..., n_channels, n_freq) if
        :code:`calc_cpsd=False`, otherwise the shape is
        (..., n_channels, n_channels, n_freq).
    coh : np.ndarray
        Coherences spectra.
        Shape is (..., n_channels, n_channels, n_freq).
        Only returned is :code:`calc_coh=True`.
    """
    n_samples, n_channels = data.shape

    if alpha is None:
        alpha = np.ones([n_samples, 1], dtype=np.float32)
    n_states = alpha.shape[-1]

    # Indices for the upper triangle of a channels by channels matrix
    m, n = np.triu_indices(n_channels)

    # Standardise before calculating spectra
    if standardize:
        data = (data - np.mean(data, axis=0)) / np.std(data, axis=0)

    # Rescaling for the PSD to account for how much time each
    # state was active. Note, we using the same scaling as the
    # the multitaper calculation in the HMM-MAR toolbox
    sum_alpha2 = np.sum(alpha**2, axis=0)
    scaling = n_samples / sum_alpha2

    psd = []
    coh = []
    for i in range(n_states):
        # Calculate spectrogram for this state's data
        x = data * alpha[:, i][..., np.newaxis]
        _, f, p = _multitaper_spectrogram(
            data=x,
            sampling_frequency=sampling_frequency,
            window_length=window_length,
            step_size=step_size,
            frequency_range=frequency_range,
            calc_cpsd=calc_cpsd or calc_coh,
            time_half_bandwidth=time_half_bandwidth,
            n_tapers=n_tapers,
        )

        # Average over the time dimension and rescale
        p = np.mean(p, axis=0)
        p *= scaling[i]
        n_freq = p.shape[-1]

        if calc_coh:
            # Create a channels by channels matrix for cross PSDs
            cpsd = np.empty(
                [n_channels, n_channels, n_freq],
                dtype=np.complex64,
            )
            cpsd[m, n] = p
            cpsd[n, m] = np.conj(p)

            # Unpack PSDs
            psd.append(cpsd[range(n_channels), range(n_channels)].real)

            # Calculate coherence
            coh.append(coherence_spectra(cpsd))

        elif calc_cpsd:
            cpsd = np.empty(
                [n_channels, n_channels, n_freq],
                dtype=np.complex64,
            )
            cpsd[m, n] = p
            cpsd[n, m] = np.conj(p)
            psd.append(cpsd)

        else:
            psd.append(p)

    # Check for nans
    if np.isnan(psd).any():
        psd = np.nan_to_num(psd)  # zero out nan values
        _logger.warn(
            "PSD contains NaN values. This may indicate a potentially "
            "poor HMM fit. You should consider running the model again "
            "or selecting the model run with the lowest free energy after "
            "training multiple times."
        )

    if not keepdims:
        # Squeeze any axes of length 1
        psd = np.squeeze(psd)
        if calc_coh:
            coh = np.squeeze(coh)

    if calc_coh:
        return f, psd, coh
    else:
        return f, psd


def _regress_welch_spectrogram(
    data,
    alpha,
    sampling_frequency,
    window_length,
    step_size,
    n_sub_windows,
    frequency_range,
    calc_cpsd,
):
    """Calculate regression spectra for a single array.

    Parameters
    ----------
    data : np.ndarray
        Data to calculate a spectrogram for.
        Shape must be (n_samples, n_channels).
    alpha : np.ndarray
        Inferred mode mixing factors. Shape must be (n_samples, n_modes).
    sampling_frequency : float
        Sampling_frequency in Hz.
    window_length : int
        Number samples to use in the window to calculate a PSD.
    step_size : int
        Step size for shifting the window.
    n_sub_windows : int, optional
        We split the window into a number of sub-windows and average the
        spectra for each sub-window. window_length must be divisible by
        :code:`n_sub_windows`.
    frequency_range : list, optional
        Minimum and maximum frequency to keep.
    calc_cpsd : bool, optional
        Should we calculate the cross spectra?

    Returns
    -------
    f : np.ndarray
        Frequency axis. Shape is (n_freq,).
    coefs : np.ndarray
        Regression coefficients. Shape is
        (n_modes, n_channels * (n_channels + 1) / 2, n_freq) if
        :code:`calc_cpsd=True`, otherwise it is (n_modes, n_channels, n_freq).
    intercept : np.ndarray
        Intercept. Shape is (n_channels * (n_channels + 1) / 2, n_freq)
        if :code:`calc_cpsd=True`, otherwise it is (n_channels, n_freq).
    """
    _, f, p = _welch_spectrogram(
        data=data,
        sampling_frequency=sampling_frequency,
        window_length=window_length,
        step_size=step_size,
        frequency_range=frequency_range,
        calc_cpsd=calc_cpsd,
        n_sub_windows=n_sub_windows,
    )
    a = _window_mean(alpha, "hann", window_length, step_size, n_sub_windows)
    coefs, intercept = regression.linear(
        a,
        p,
        fit_intercept=True,
        normalize=True,
        log_message=False,
    )
    return f, coefs, intercept


def _window_mean(alpha, window, window_length, step_size, n_sub_windows):
    """Applies a windowing function to a time series and takes the mean.

    Parameters
    ----------
    alpha : np.ndarray
        Time series data. Shape is (n_samples, n_modes).
    window : str
        Name of the windowing function.
    window_length : int
        Number of data points in a window.
    step_size : int
        Step size for shifting the window.
    n_sub_windows : int
        We split the window into a number of sub-windows and average
        across all sub-windows. This is the number of sub-windows.

    Returns
    -------
    mean_window_alpha : np.ndarray
        Mean for each window.
    """
    n_samples, n_modes = alpha.shape

    # Pad alphas
    alpha = np.pad(alpha, window_length // 2)[
        :, window_length // 2 : window_length // 2 + n_modes
    ]

    # Window to apply to alpha
    window = signal.get_window(window, window_length // n_sub_windows)

    # Indices of time points to calculate a periodogram for
    time_indices = range(0, n_samples, step_size)
    n_windows = n_samples // step_size

    # Array to hold mean of alpha multiplied by the windowing function
    mean_window_alpha = np.empty([n_windows, n_modes], dtype=np.float32)
    for i in range(n_windows):
        # Alpha in the window
        j = time_indices[i]
        a_window = alpha[j : j + window_length]

        # Calculate alpha for the sub-window by taking the mean
        # over time after applying the windowing function
        a_sub_window = np.empty([n_sub_windows, n_modes], dtype=np.float32)
        for k in range(n_sub_windows):
            a_sub_window[k] = np.mean(
                a_window[
                    k
                    * window_length
                    // n_sub_windows : (k + 1)
                    * window_length
                    // n_sub_windows
                ]
                * window[..., np.newaxis],
                axis=0,
            )

        # Average alpha for each sub-window
        mean_window_alpha[i] = np.mean(a_sub_window, axis=0)

    return mean_window_alpha
